import os
import pytest
import unittest
from typing import List

from sacrerouge.data import EvalInstance, Metrics
from sacrerouge.data.dataset_readers import ReferenceBasedDatasetReader
from sacrerouge.io import JsonlReader
from sacrerouge.metrics import Rouge
from sacrerouge.commands.score import score_instances

_task1_summaries_file_path = 'datasets/duc-tac/duc2007/v1.0/task1.summaries.jsonl'
_task1_metrics_file_path = 'datasets/duc-tac/duc2007/v1.0/task1.metrics.jsonl'

_task2_summaries_file_path = 'datasets/duc-tac/duc2007/v1.0/task2.A-B-C.summaries.jsonl'
_task2_metrics_file_path = 'datasets/duc-tac/duc2007/v1.0/task2.A-B-C.metrics.jsonl'


class TestDUC2007Rouge(unittest.TestCase):
    def _filter_by_type(self,
                        instances: List[EvalInstance],
                        metrics_list: List[Metrics],
                        summarizer_type: str):
        filtered_instances = []
        filtered_metrics = []
        for instance, metrics in zip(instances, metrics_list):
            if instance.summarizer_type == summarizer_type:
                filtered_instances.append(instance)
                filtered_metrics.append(metrics)
        return filtered_instances, filtered_metrics

    @pytest.mark.skipif(not os.path.exists(_task1_summaries_file_path), reason='DUC 2007 task 1 summaries file does not exist')
    @pytest.mark.skipif(not os.path.exists(_task1_metrics_file_path), reason='DUC 2007 task 1 metrics file does not exist')
    def test_task1_rouge(self):
        # Test the first several instances in the DUC 2007 data to ensure that
        # our computation of ROUGE matches the values released by NIST. In this year,
        # they did not release the raw output for the non-jackknifing runs, so
        # we cannot compare those scores
        instances = ReferenceBasedDatasetReader().read(_task1_summaries_file_path)
        metrics_list = JsonlReader(_task1_metrics_file_path, Metrics).read()
        metric_names = ['rouge-1', 'rouge-2', 'rouge-3', 'rouge-4', 'rouge-l', 'rouge-su4', 'rouge-w-1.2']
        rouge = Rouge(max_ngram=4,
                      use_porter_stemmer=True,
                      remove_stopwords=False,
                      compute_rouge_l=True,
                      skip_bigram_gap_length=4,
                      wlcs_weight=1.2)

        peer_instances, peer_metrics = self._filter_by_type(instances, metrics_list, 'peer')
        reference_instances, reference_metrics = self._filter_by_type(instances, metrics_list, 'reference')

        num_to_check = 25
        actual_metrics_dicts = score_instances(peer_instances[:num_to_check], [rouge])
        for expected_metrics in peer_metrics[:num_to_check]:
            instance_id = expected_metrics.instance_id
            summarizer_id = expected_metrics.summarizer_id
            actual_metrics = actual_metrics_dicts[instance_id][summarizer_id]

            for metric in metric_names:
                assert actual_metrics.metrics[metric + '_jk'] == expected_metrics.metrics[metric + '_jk']

        actual_metrics_dicts = score_instances(reference_instances[:num_to_check], [rouge])
        for expected_metrics in reference_metrics[:num_to_check]:
            instance_id = expected_metrics.instance_id
            summarizer_id = expected_metrics.summarizer_id
            actual_metrics = actual_metrics_dicts[instance_id][summarizer_id]

            for metric in metric_names:
                assert actual_metrics.metrics[metric + '_jk'] == expected_metrics.metrics[metric + '_jk']

    @pytest.mark.skipif(not os.path.exists(_task2_summaries_file_path), reason='DUC 2007 task 2 summaries file does not exist')
    @pytest.mark.skipif(not os.path.exists(_task2_metrics_file_path), reason='DUC 2007 task 2 metrics file does not exist')
    def test_task2_rouge(self):
        # Test the first several instances in the DUC 2007 data to ensure that
        # our computation of ROUGE matches the values released by NIST. In this year,
        # they did not release the raw output for the non-jackknifing runs, so
        # we cannot compare those scores
        instances = ReferenceBasedDatasetReader().read(_task2_summaries_file_path)
        metrics_list = JsonlReader(_task2_metrics_file_path, Metrics).read()
        metric_names = ['rouge-1', 'rouge-2', 'rouge-3', 'rouge-4', 'rouge-l', 'rouge-su4', 'rouge-w-1.2']
        rouge = Rouge(max_ngram=4,
                      use_porter_stemmer=True,
                      remove_stopwords=False,
                      compute_rouge_l=True,
                      skip_bigram_gap_length=4,
                      wlcs_weight=1.2)

        peer_instances, peer_metrics = self._filter_by_type(instances, metrics_list, 'peer')
        reference_instances, reference_metrics = self._filter_by_type(instances, metrics_list, 'reference')

        num_to_check = 25
        actual_metrics_dicts = score_instances(peer_instances[:num_to_check], [rouge])
        for expected_metrics in peer_metrics[:num_to_check]:
            instance_id = expected_metrics.instance_id
            summarizer_id = expected_metrics.summarizer_id
            actual_metrics = actual_metrics_dicts[instance_id][summarizer_id]

            for metric in metric_names:
                assert actual_metrics.metrics[metric + '_jk'] == pytest.approx(expected_metrics.metrics[metric + '_jk'], abs=1e-3)

        actual_metrics_dicts = score_instances(reference_instances[:num_to_check], [rouge])
        for expected_metrics in reference_metrics[:num_to_check]:
            instance_id = expected_metrics.instance_id
            summarizer_id = expected_metrics.summarizer_id
            actual_metrics = actual_metrics_dicts[instance_id][summarizer_id]

            for metric in metric_names:
                assert actual_metrics.metrics[metric + '_jk'] == expected_metrics.metrics[metric + '_jk']



import csv
import jsons
import zipfile
from collections import defaultdict

from sacrerouge.data import Metrics, MetricsDict
from sacrerouge.io import JsonlWriter
from sacrerouge.datasets.multiling.util import LANGUAGE_CODES


def load_model_summaries(summaries_zip: str):
    summaries = defaultdict(lambda: defaultdict(dict))
    with zipfile.ZipFile(summaries_zip, 'r') as zip:
        for file_path in zip.namelist():
            if file_path.endswith('.250'):
                path = file_path.split('/')
                language_code = LANGUAGE_CODES[path[1]]
                parts = path[2].split('.')
                instance_id = parts[0]
                annotator = parts[1]

                # The summaries don't seem to have any structure, so
                # just read in the full text.
                contents = zip.open(file_path, 'r').read().decode()
                text = list(filter(None, [line.strip() for line in contents.splitlines()]))
                summaries[language_code][instance_id][annotator] = {
                    'summarizer_id': annotator,
                    'summarizer_type': 'reference',
                    'text': text
                }
    return summaries


def load_peer_summaries(summaries_zip: str):
    summaries = defaultdict(lambda: defaultdict(dict))
    with zipfile.ZipFile(summaries_zip, 'r') as zip:
        for file_path in zip.namelist():
            if file_path.endswith('.250'):
                path = file_path.split('/')
                summarizer_id = path[1]
                language_code = LANGUAGE_CODES[path[2]]

                parts = path[3].split('.')
                instance_id = parts[0]

                contents = zip.open(file_path, 'r').read().decode()
                text = list(filter(None, [line.strip() for line in contents.splitlines()]))

                summaries[language_code][instance_id][summarizer_id] = {
                    'summarizer_id': summarizer_id,
                    'summarizer_type': 'reference',
                    'text': text
                }
    return summaries


def load_metrics(grades_zip: str):
    metrics = defaultdict(lambda: defaultdict(lambda: defaultdict(MetricsDict)))
    with zipfile.ZipFile(grades_zip, 'r') as zip:
        for file_path in zip.namelist():
            if file_path.endswith('.csv'):
                reader = csv.reader(zip.open(file_path, 'r').read().decode().splitlines(), delimiter='\t')
                for i, line in enumerate(reader):
                    if i == 0:
                        continue  # header
                    instance_id, summarizer_id, language, _, _, overall_responsiveness = line
                    language_code = LANGUAGE_CODES[language]
                    overall_responsiveness = float(overall_responsiveness)
                    if 'overall_responsiveness' not in metrics[language_code][instance_id][summarizer_id]:
                        metrics[language_code][instance_id][summarizer_id] = MetricsDict({'overall_responsiveness': []})
                    metrics[language_code][instance_id][summarizer_id]['overall_responsiveness'].append(overall_responsiveness)
    return metrics


def save_data(model_summaries, peer_summaries, metrics, output_dir):
    for language in model_summaries.keys():
        with JsonlWriter(f'{output_dir}/{language}.summaries.jsonl') as out_summaries:
            if language in metrics:
                out_metrics = open(f'{output_dir}/{language}.metrics.jsonl', 'w')
            else:
                out_metrics = None

            for instance_id in sorted(model_summaries[language].keys()):
                references = list(model_summaries[language][instance_id][key] for key in sorted(model_summaries[language][instance_id].keys()))
                for summarizer_id in sorted(peer_summaries[language][instance_id].keys()):
                    summary = peer_summaries[language][instance_id][summarizer_id]
                    out_summaries.write({
                        'instance_id': instance_id,
                        'summarizer_id': summarizer_id,
                        'summarizer_type': 'peer',
                        'summary': summary,
                        'references': references
                    })
                    if out_metrics is not None:
                        instance_metrics = metrics[language][instance_id][summarizer_id]
                        out_metrics.write(jsons.dumps(Metrics(instance_id, summarizer_id, 'peer', instance_metrics)) + '\n')

                for i, reference in enumerate(references):
                    summarizer_id = reference['summarizer_id']
                    out_summaries.write({
                        'instance_id': instance_id,
                        'summarizer_id': summarizer_id,
                        'summarizer_type': 'reference',
                        'summary': reference,
                        'references': references[:i] + references[i + 1:]
                    })
                    if out_metrics is not None:
                        instance_metrics = metrics[language][instance_id][summarizer_id]
                        out_metrics.write(jsons.dumps(Metrics(instance_id, summarizer_id, 'reference', instance_metrics)) + '\n')

            if out_metrics is not None:
                out_metrics.close()


def setup(model_summaries_zip: str, peer_summaries_zip: str, grades_zip: str, output_dir: str) -> None:
    model_summaries = load_model_summaries(model_summaries_zip)
    peer_summaries = load_peer_summaries(peer_summaries_zip)
    metrics = load_metrics(grades_zip)
    save_data(model_summaries, peer_summaries, metrics, output_dir)


import json

from sacrerouge.common.util import download_url_to_file
from sacrerouge.data import Metrics, MetricsDict
from sacrerouge.io import JsonlWriter


def _download_raw_data(output_dir: str, force: bool) -> None:
    download_url_to_file('https://github.com/PrimerAI/blanc/raw/master/data/CNN_DailyMail_555.json',
                         f'{output_dir}/raw/CNN_DailyMail_555.json', force=force)
    download_url_to_file('https://github.com/PrimerAI/blanc/raw/master/data/DailyNews_300.json',
                         f'{output_dir}/raw/DailyNews_300.json', force=force)
    download_url_to_file('https://github.com/PrimerAI/blanc/raw/master/data/DailyNews_300_aspects.json',
                         f'{output_dir}/raw/DailyNews_300_aspects.json', force=force)


def _load_generic_scores(input_file: str):
    data = json.load(open(input_file, 'r'))
    instances = []
    metrics_list = []
    documents = {}
    for i, instance in enumerate(data):
        document = instance['text'].strip()
        summary = instance['summary'].strip()
        scores = instance['scores']
        summarizer_id = str(i)

        if document not in documents:
            documents[document] = str(len(documents))
        instance_id = documents[document]

        instances.append({
            'instance_id': instance_id,
            'summarizer_id': summarizer_id,
            'summarizer_type': 'peer',
            'summary': {'text': summary},
            'document': {'text': document}
        })
        metrics_list.append(Metrics(instance_id, summarizer_id, 'peer', MetricsDict({'generic_quality': scores})))
    return instances, metrics_list


def _load_dailynews_aspects(input_file: str):
    data = json.load(open(input_file, 'r'))
    instances = []
    metrics_list = []
    documents = {}
    for i, instance in enumerate(data):
        document = instance['text'].strip()
        summary = instance['summary'].strip()
        fluent = instance['map_quality_scores']['Fluent']['scores']
        understandable = instance['map_quality_scores']['Understandable']['scores']
        informative = instance['map_quality_scores']['Informative']['scores']
        compact = instance['map_quality_scores']['Compact']['scores']
        overall = instance['map_quality_scores']['Overall']['scores']
        summarizer_id = str(i)

        if document not in documents:
            documents[document] = str(len(documents))
        instance_id = documents[document]

        instances.append({
            'instance_id': instance_id,
            'summarizer_id': summarizer_id,
            'summarizer_type': 'peer',
            'summary': {'text': summary},
            'document': {'text': document}
        })
        metrics_list.append(Metrics(instance_id, summarizer_id, 'peer', MetricsDict({
            'fluent': fluent,
            'understandable': understandable,
            'informative': informative,
            'compact': compact,
            'overall': overall,
        })))
    return instances, metrics_list


def _save(instances, output_file):
    with JsonlWriter(output_file) as out:
        for item in instances:
            out.write(item)


def setup(output_dir: str, force: bool) -> None:
    _download_raw_data(output_dir, force)
    cnn_dm_summaries, cnn_dm_metrics_list = _load_generic_scores(f'{output_dir}/raw/CNN_DailyMail_555.json')
    dailynews_summaries, dailynews_metrics_list = _load_generic_scores(f'{output_dir}/raw/DailyNews_300.json')
    dailynews_aspects_summaries, dailynews_aspects_metrics_list = _load_dailynews_aspects(f'{output_dir}/raw/DailyNews_300_aspects.json')

    _save(cnn_dm_summaries, f'{output_dir}/cnn-dailymail.summaries.jsonl')
    _save(cnn_dm_metrics_list, f'{output_dir}/cnn-dailymail.metrics.jsonl')

    _save(dailynews_summaries, f'{output_dir}/dailynews.summaries.jsonl')
    _save(dailynews_metrics_list, f'{output_dir}/dailynews.metrics.jsonl')

    _save(dailynews_aspects_summaries, f'{output_dir}/dailynews-aspects.summaries.jsonl')
    _save(dailynews_aspects_metrics_list, f'{output_dir}/dailynews-aspects.metrics.jsonl')



import argparse
import lxml.html
import os
import re
import tarfile
from collections import defaultdict
from io import BytesIO
from nltk.tokenize import sent_tokenize
from typing import Any, Dict, List

from sacrerouge.io import JsonlWriter


def load_task1_and_task2_documents(documents_tar_path: str):
    cluster_to_doc_ids = defaultdict(list)
    documents = {}

    with tarfile.open(documents_tar_path, 'r') as tar:
        for member in tar.getmembers():
            if member.name.startswith('DUC2004_Summarization_Documents/duc2004_testdata/tasks1and2/duc2004_tasks1and2_docs/docs'):
                if member.isfile():
                    path = member.name.split('/')
                    cluster = path[-2][:-1]
                    filename = path[-1]
                    cluster_to_doc_ids[cluster].append(filename)

                    date = filename[3:11]
                    year, month, day = date[:4], date[4:6], date[6:]
                    date = f'{year}-{month}-{day}'

                    html = tar.extractfile(member).read().decode()
                    tree = lxml.html.document_fromstring(html)
                    text_nodes = list(tree.xpath('//text'))
                    assert len(text_nodes) == 1

                    text = text_nodes[0].text_content().strip()
                    text = re.sub('\s+', ' ', text)
                    sentences = sent_tokenize(text)

                    document = {
                        'filename': filename,
                        'date': date,
                        'text': sentences
                    }
                    documents[filename] = document

    print(f'Loaded {len(cluster_to_doc_ids)} clusters and {len(documents)} documents for tasks 1 and 2')
    return cluster_to_doc_ids, documents


def load_task5_documents(documents_tar_path: str):
    cluster_to_doc_ids = defaultdict(list)
    documents = {}

    with tarfile.open(documents_tar_path, 'r') as tar:
        for member in tar.getmembers():
            if member.name.startswith('DUC2004_Summarization_Documents/duc2004_testdata/task5/duc2004_task5_docs/docs'):
                if member.isfile():
                    path = member.name.split('/')
                    cluster = path[-2][:-1]
                    filename = path[-1]
                    cluster_to_doc_ids[cluster].append(filename)

                    date = filename[3:11]
                    year, month, day = date[:4], date[4:6], date[6:]
                    date = f'{year}-{month}-{day}'

                    html = tar.extractfile(member).read().decode()
                    tree = lxml.html.document_fromstring(html)
                    text_nodes = list(tree.xpath('//text'))
                    assert len(text_nodes) == 1

                    text = text_nodes[0].text_content().strip()
                    text = re.sub('\s+', ' ', text)
                    sentences = sent_tokenize(text)

                    document = {
                        'filename': filename,
                        'date': date,
                        'text': sentences
                    }
                    documents[filename] = document

    print(f'Loaded {len(cluster_to_doc_ids)} clusters and {len(documents)} documents for task 5')
    return cluster_to_doc_ids, documents


def load_task5_topics(documents_tar_path: str):
    topics = {}
    with tarfile.open(documents_tar_path, 'r') as tar:
        member = tar.extractfile('DUC2004_Summarization_Documents/duc2004_testdata/task5/duc2004.task5.topicsets')
        lines = member.read().decode().splitlines()
        for line in lines[7:]:
            cluster, topic = line.strip().split('\t')
            cluster = cluster[:-1]
            topics[cluster] = topic

    print(f'Loaded {len(topics)} topics')
    return topics


def load_task1_summaries(results_tar: str):
    summaries = defaultdict(list)
    with tarfile.open(results_tar, 'r') as tar:
        inner_tar_bytes = tar.extractfile('duc2004_results/ROUGE/duc2004.task1.ROUGE.models.tar.gz').read()
        with tarfile.open(fileobj=BytesIO(inner_tar_bytes)) as inner_tar:
            for member in inner_tar.getmembers():
                filename = os.path.basename(member.name)
                parts = filename.split('.')
                filename = parts[-2] + '.' + parts[-1]

                text = [inner_tar.extractfile(member).read().decode().strip()]
                summary = {
                    'text': text
                }

                if summary not in summaries[filename]:
                    summaries[filename].append(summary)
                else:
                    print(f'Skipping duplicate summary for document {filename}')

    num_summaries = sum(len(s) for s in summaries.values())
    print(f'Loaded {num_summaries} summaries for task 1')
    return summaries


def load_task2_summaries(results_tar: str):
    summaries = defaultdict(list)
    with tarfile.open(results_tar, 'r') as tar:
        inner_tar_bytes = tar.extractfile('duc2004_results/ROUGE/duc2004.task2.ROUGE.models.tar.gz').read()
        with tarfile.open(fileobj=BytesIO(inner_tar_bytes)) as inner_tar:
            for member in inner_tar.getmembers():
                filename = os.path.basename(member.name)
                parts = filename.split('.')
                cluster = parts[0].lower()

                sentences = inner_tar.extractfile(member).read().decode().splitlines()
                summary = {
                    'text': sentences
                }
                if summary not in summaries[cluster]:
                    summaries[cluster].append(summary)
                else:
                    print(f'Skipping duplicate summary for cluster {cluster}')

    num_summaries = sum(len(s) for s in summaries.values())
    print(f'Loaded {num_summaries} summaries for task 2')
    return summaries


def load_task5_summaries(results_tar: str):
    summaries = defaultdict(list)
    with tarfile.open(results_tar, 'r') as tar:
        inner_tar_bytes = tar.extractfile('duc2004_results/ROUGE/duc2004.task5.ROUGE.models.tar.gz').read()
        with tarfile.open(fileobj=BytesIO(inner_tar_bytes)) as inner_tar:
            for member in inner_tar.getmembers():
                filename = os.path.basename(member.name)
                parts = filename.split('.')
                cluster = parts[0].lower()

                sentences = inner_tar.extractfile(member).read().decode().splitlines()
                summary = {
                    'text': sentences
                }
                if summary not in summaries[cluster]:
                    summaries[cluster].append(summary)
                else:
                    print(f'Skipping duplicate summary for cluster {cluster}')

    num_summaries = sum(len(s) for s in summaries.values())
    print(f'Loaded {num_summaries} summaries for task 5')
    return summaries


def save_task1_data(documents: Dict[str, List[str]],
                    summaries: Dict[str, List[List[str]]],
                    file_path: str):
    with JsonlWriter(file_path) as out:
        for doc_id, document in documents.items():
            doc_summaries = summaries[doc_id]
            data = {
                'instance_id': doc_id,
                'document': document,
                'summaries': doc_summaries
            }
            if len(doc_summaries) != 4:
                print(f'Task 1 document {doc_id} has {len(doc_summaries)} summaries. Expected 4')
            out.write(data)


def save_task2_data(cluster_to_doc_ids: Dict[str, List[str]],
                    documents: Dict[str, List[str]],
                    summaries: Dict[str, List[List[str]]],
                    file_path: str):
    with JsonlWriter(file_path) as out:
        for cluster, doc_ids in cluster_to_doc_ids.items():
            cluster_docs = [documents[doc_id] for doc_id in doc_ids]
            cluster_summaries = summaries[cluster]
            data = {
                'instance_id': cluster,
                'documents': cluster_docs,
                'summaries': cluster_summaries
            }
            if len(cluster_summaries) != 4:
                print(f'Task 2 cluster {cluster} has {len(cluster_summaries)} summaries. Expected 4')
            out.write(data)


def save_task5_data(cluster_to_doc_ids: Dict[str, List[str]],
                    documents: Dict[str, List[str]],
                    summaries: Dict[str, List[List[str]]],
                    topics: Dict[str, str],
                    file_path: str):
    with JsonlWriter(file_path) as out:
        for cluster, doc_ids in cluster_to_doc_ids.items():
            cluster_docs = [documents[doc_id] for doc_id in doc_ids]
            cluster_summaries = summaries[cluster]
            cluster_topic = topics[cluster]
            data = {
                'instance_id': cluster,
                'topic': cluster_topic,
                'documents': cluster_docs,
                'summaries': cluster_summaries
            }
            if len(cluster_summaries) != 4:
                print(f'Task 5 cluster {cluster} has {len(cluster_summaries)} summaries. Expected 4')
            out.write(data)


def setup(data_root: str, output_dir: str) -> None:
    documents_tar = f'{data_root}/from-nist/DUC2004_Summarization_Documents.tgz'
    results_tar = f'{data_root}/scrapes/duc.nist.gov/past_duc/duc2004/duc2004_results.tgz'
    main(documents_tar, results_tar, output_dir)


def main(documents_tar, results_tar, output_dir):
    cluster_to_doc_ids, documents = load_task1_and_task2_documents(documents_tar)
    task_1_summaries = load_task1_summaries(results_tar)
    save_task1_data(documents, task_1_summaries, f'{output_dir}/task1.jsonl')

    task_2_summaries = load_task2_summaries(results_tar)
    save_task2_data(cluster_to_doc_ids, documents, task_2_summaries, f'{output_dir}/task2.jsonl')

    cluster_to_doc_ids, documents = load_task5_documents(documents_tar)
    topics = load_task5_topics(documents_tar)
    task_5_summaries = load_task5_summaries(results_tar)
    save_task5_data(cluster_to_doc_ids, documents, task_5_summaries, topics, f'{output_dir}/task5.jsonl')


if __name__ == '__main__':
    argp = argparse.ArgumentParser()
    argp.add_argument('documents_tar')
    argp.add_argument('results_tar')
    argp.add_argument('topics_file_path')
    argp.add_argument('output_dir')
    args = argp.parse_args()

    main(args.documents_tar, args.results_tar, args.topics_file_path, args.output_dir)



import tarfile
from collections import defaultdict
from nltk.tokenize import sent_tokenize

from sacrerouge.io import JsonlWriter


def load_documents(tar_path: str):
    documents = defaultdict(dict)
    with tarfile.open(tar_path, 'r') as tar:
        for member in tar.getmembers():
            if member.name.startswith('2017_test_data/body/text') and member.name.endswith('.txt'):
                if 'simple' in member.name:
                    # Seems to be a dummy language
                    continue
                language = member.name[25:27]
                instance_id = member.name[28:60]
                text = tar.extractfile(member).read().decode().splitlines()
                if language == 'en':
                    text = [sent_tokenize(paragraph) for paragraph in text]
                document = {
                    'filename': instance_id,
                    'text': text
                }
                documents[language][instance_id] = document
    return documents


def load_summaries(eval_path: str):
    summaries = defaultdict(dict)
    with tarfile.open(eval_path, 'r') as tar:
        for member in tar.getmembers():
            if member.name.startswith('Evaluation/2017_summaries/') and member.name.endswith('.txt'):
                if 'simple' in member.name:
                    # Seems to be a dummy language
                    continue
                path = member.name.split('/')
                language_code = path[2]
                instance_id = path[3][:32]
                text = tar.extractfile(member).read().decode().splitlines()
                if language_code == 'en':
                    text = [sent_tokenize(paragraph) for paragraph in text]
                summaries[language_code][instance_id] = {
                    'text': text
                }
    return summaries


def save_data(documents, summaries, output_dir: str):
    for language in documents.keys():
        with JsonlWriter(f'{output_dir}/{language}.jsonl') as out:
            for instance_id in sorted(documents[language].keys()):
                out.write({
                    'instance_id': instance_id,
                    'document': documents[language][instance_id],
                    'summary': summaries[language][instance_id]
                })


def setup(data_path: str, eval_path: str, output_dir: str):
    documents = load_documents(data_path)
    summaries = load_summaries(eval_path)
    save_data(documents, summaries, output_dir)



import jsons
import unittest

from sacrerouge.data import Metrics, MetricsDict


class TestMetrics(unittest.TestCase):
    def test_serialization(self):
        instance_id = 'd500'
        summarizer_id = '5'
        summarizer_type = 'peer'
        metrics_dict = {'a': 4, 'b': {'c': [1, 2]}}
        metrics = Metrics(instance_id, summarizer_id, summarizer_type, metrics_dict)

        serialized = jsons.dumps(metrics)
        assert serialized == '{"instance_id": "d500", "summarizer_id": "5", "summarizer_type": "peer", "metrics": {"a": 4, "b": {"c": [1, 2]}}}'

        deserialized = jsons.loads(serialized, Metrics)
        assert metrics == deserialized
        assert isinstance(deserialized.metrics, MetricsDict)



import argparse
import hashlib
import logging
import os
import shutil
import zipfile
from overrides import overrides
from packaging import version
from tqdm import tqdm
from typing import Any, Dict, List, Tuple

from sacrerouge.commands import MetricSetupSubcommand
from sacrerouge.common import DATA_ROOT
from sacrerouge.common.util import download_file_from_google_drive
from sacrerouge.data import MetricsDict
from sacrerouge.data.types import ReferenceType, SummaryType
from sacrerouge.metrics import Metric, ReferenceBasedMetric

MIN_QAEVAL_VERSION = '0.0.8'
QAEVAL_INSTALLED = False

logger = logging.getLogger(__name__)

try:
    import qaeval
    installed_version = version.parse(qaeval.__version__)
    if installed_version < version.parse(MIN_QAEVAL_VERSION):
        raise ImportError(f'"qaeval" version out of date. Minimum required: {MIN_QAEVAL_VERSION}')

except ImportError:
    @Metric.register('qa-eval')
    class QAEval(ReferenceBasedMetric):
        def __init__(self) -> None:
            super().__init__()

        def score_multi_all(self,
                            summaries_list: List[List[SummaryType]],
                            references_list: List[List[ReferenceType]]) -> List[List[MetricsDict]]:
            raise Exception(
                f'Package "qaeval" could not be imported. Please install "qaeval" before running the metric. '
                f'The minimum required version is {MIN_QAEVAL_VERSION}')
else:
    from qaeval import AnswerSelector, QuestionAnsweringModel, QuestionGenerationModel
    from qaeval.answer_selection import NP_CHUNKS_STRATEGY, AnswerOffsets
    from qaeval.scoring.scorers import ExactMatchScorer, F1Scorer, IsAnsweredScorer, LERCScorer, MetaScorer

    QAEVAL_INSTALLED = True

    @Metric.register('qa-eval')
    class QAEval(ReferenceBasedMetric):
        def __init__(self,
                     answer_selection_strategy: str = NP_CHUNKS_STRATEGY,
                     generation_model_path: str = f'{DATA_ROOT}/metrics/qaeval/models/generation/model.tar.gz',
                     answering_model_dir: str = f'{DATA_ROOT}/metrics/qaeval/models/answering/model',
                     cuda_device: int = 0,
                     generation_batch_size: int = 8,
                     answering_batch_size: int = 8,
                     use_lerc: bool = False,
                     lerc_model_path: str = f'{DATA_ROOT}/metrics/qaeval/models/lerc/model.tar.gz',
                     lerc_pretrained_model_path: str = f'{DATA_ROOT}/metrics/qaeval/models/lerc/pretrained.tar.gz',
                     lerc_batch_size: int = 8,
                     verbose: bool = False) -> None:
            super().__init__()
            self.answer_selector = AnswerSelector(answer_selection_strategy)
            self.question_generator = QuestionGenerationModel(generation_model_path, cuda_device=cuda_device,
                                                              batch_size=generation_batch_size, silent=not verbose)
            self.question_answerer = QuestionAnsweringModel(answering_model_dir, cuda_device=cuda_device,
                                                            batch_size=answering_batch_size, silent=not verbose)
            self.verbose = verbose

            scorers = [IsAnsweredScorer(), ExactMatchScorer(), F1Scorer()]
            if use_lerc:
                scorers.append(LERCScorer(lerc_model_path, lerc_pretrained_model_path, cuda_device, lerc_batch_size))
            self.scorer = MetaScorer(scorers)

        def _flatten_summaries(self, summaries_list: List[List[SummaryType]]) -> List[List[str]]:
            # Flattens all of the summaries so they are `str` instead of `List[str]`
            flat_summaries_list = []
            for summaries in summaries_list:
                flat_summaries_list.append([])
                for summary in summaries:
                    if isinstance(summary, list):
                        summary = ' '.join(summary)
                    flat_summaries_list[-1].append(summary)
            return flat_summaries_list

        def _unroll_summaries(self,
                              summaries_list: List[List[str]],
                              references_list: List[List[str]]) -> Tuple[List[str], List[List[str]]]:
            # The summaries are originally grouped by identical references. This will flatten the list of summaries
            # and copy the references into a parallel list for easier processing.
            unrolled_summaries = []
            unrolled_references = []
            for summaries, references in zip(summaries_list, references_list):
                for summary in summaries:
                    unrolled_summaries.append(summary)
                    unrolled_references.append(references)
            return unrolled_summaries, unrolled_references

        def _get_empty_summary_mask(self,
                                    summaries: List[str],
                                    references_list: List[List[str]]) -> Tuple[List[str], List[List[str]], List[bool]]:
            # This will identify any summaries that have empty text. The output will be the list of non-empty summaries
            # with their corresponding references plus a list of booleans that is parallel will the input `summaries`
            # which mark whether or not they are empty
            is_empty_list = []
            non_empty_summaries = []
            non_empty_references_list = []

            for summary, references in zip(summaries, references_list):
                if len(summary.strip()) > 0:
                    is_empty_list.append(False)
                    non_empty_summaries.append(summary)
                    non_empty_references_list.append(references)
                else:
                    is_empty_list.append(True)
            return non_empty_summaries, non_empty_references_list, is_empty_list

        def _get_question_id(self, instance_index: int, reference_index: int, start: int, end: int) -> str:
            m = hashlib.md5()
            m.update(str(instance_index).encode())
            m.update(str(reference_index).encode())
            m.update(str(start).encode())
            m.update(str(end).encode())
            return m.hexdigest()

        def _generate_qa_pairs(self, references_list: List[List[str]]) -> List[List[List[Dict[str, Any]]]]:
            # This will generate the question-answer pairs for each reference. Since references may be repeated,
            # we first deduplicate the references to minimize the expensive work.
            #
            # `reference_to_index` keeps track of where each of the unique references are in `distinct_references_list`
            reference_to_index = {}
            distinct_references_list = []

            # Maps from (i, j) to the index in the `distinct_references_list`
            mapping = {}
            for i, references in enumerate(references_list):
                for j, reference in enumerate(references):
                    if reference not in reference_to_index:
                        reference_to_index[reference] = len(distinct_references_list)
                        distinct_references_list.append(reference)
                    mapping[(i, j)] = reference_to_index[reference]

            # Select the answers
            logger.info(f'Selecting answers from {len(distinct_references_list)} distinct summaries')
            answers_list = self.answer_selector.select_all(distinct_references_list)
            num_answers = sum(len(answers) for answers in answers_list)
            logger.info(f'Selected {num_answers} answers in total')

            # Generate the questions
            generation_inputs = []
            for reference, answers in zip(distinct_references_list, answers_list):
                for answer in answers:
                    sentence = reference[answer.sent_start:answer.sent_end]
                    start = answer.start - answer.sent_start
                    end = answer.end - answer.sent_start
                    generation_inputs.append((sentence, start, end))

            logger.info(f'Generating questions for {len(generation_inputs)} answers')
            question_list = self.question_generator.generate_all(generation_inputs)
            logger.info('Finished generating questions')

            # Remap the questions to align with the answers
            index = 0
            remapped_questions = []
            for i, answers in enumerate(answers_list):
                remapped_questions.append([])
                for _ in answers:
                    remapped_questions[-1].append(question_list[index])
                    index += 1
                assert len(remapped_questions[i]) == len(answers_list[i])
            assert len(remapped_questions) == len(answers_list)

            # Remap output to align with the inputs
            # qa_pairs_lists[summary_index][reference_index] = [(q, a)]
            qa_pairs_lists = []
            for i, references in enumerate(references_list):
                qa_pairs_lists.append([])
                for j, reference in enumerate(references):
                    index = mapping[(i, j)]
                    qa_pairs_lists[-1].append([])
                    for question, answer in zip(remapped_questions[index], answers_list[index]):
                        question_id = self._get_question_id(i, j, answer.start, answer.end)
                        qa_pairs_lists[-1][-1].append({
                            'question_id': question_id,
                            'question': question,
                            'answer': answer.text,
                            'sent_start': answer.sent_start,
                            'sent_end': answer.sent_end,
                            'answer_start': answer.start,
                            'answer_end': answer.end
                        })
            return qa_pairs_lists

        def _get_prediction_id(self, prediction_index: int):
            m = hashlib.md5()
            m.update(str(prediction_index).encode())
            return m.hexdigest()

        def _answer_questions(self,
                              summaries: List[str],
                              qa_pairs_lists: List[List[List[Dict[str, Any]]]]) -> List[List[List[Dict[str, Any]]]]:
            # Answers all of the questions. Some of the (question, context) pairs may be duplicates, for instance because
            # of jackknifing. It will be a lot faster to deduplicate them first.
            #
            # `qa_inputs` will contain the unique inputs, `context_to_input_index` maps from the (question, context) pair
            # to its index in `qa_inputs`, and `mapping` will map from the i-th summary, j-th reference, and k-th question
            # to the index of the corresponding data in `qa_inputs`
            qa_inputs = []
            context_to_input_index = {}
            mapping = {}

            for i, (summary, qa_pairs_list) in enumerate(zip(summaries, qa_pairs_lists)):
                for j, qa_pairs in enumerate(qa_pairs_list):
                    for k, qa in enumerate(qa_pairs):
                        question = qa['question']
                        key = (question, summary)
                        if key not in context_to_input_index:
                            context_to_input_index[key] = len(qa_inputs)
                            qa_inputs.append(key)
                        mapping[(i, j, k)] = context_to_input_index[key]

            logger.info(f'Answering {len(qa_inputs)} distinct (question, context) pairs')
            predictions = self.question_answerer.answer_all(qa_inputs, return_offsets=True)
            logger.info('Finished answering questions')

            # Remap from the distinct answers back to the original QA lists
            predictions_lists = []
            for i, (summary, qa_pairs_list) in enumerate(zip(summaries, qa_pairs_lists)):
                predictions_lists.append([])
                for j, qa_pairs in enumerate(qa_pairs_list):
                    predictions_lists[-1].append([])
                    for k, qa in enumerate(qa_pairs):
                        index = mapping[(i, j, k)]
                        prediction, probability, null_probability, offsets = predictions[index]
                        predictions_lists[-1][-1].append({
                            'prediction_id': self._get_prediction_id(index),
                            'prediction': prediction,
                            'probability': probability,
                            'null_probability': null_probability,
                            'start': offsets[0],
                            'end': offsets[1],
                        })
            return predictions_lists

        def _score_predictions(self,
                               summaries: List[str],
                               qa_pairs_lists: List[List[List[Dict[str, Any]]]],
                               predictions_lists: List[List[List[Dict[str, Any]]]]) -> Tuple[List[MetricsDict], List[List[List[Dict[str, float]]]]]:
            logger.info('Scoring predictions')
            metrics_list = []
            scores_list = []

            generator = tqdm(
                zip(summaries, qa_pairs_lists, predictions_lists),
                total=len(summaries),
                disable=not self.verbose
            )
            for summary, qa_pairs_list, predictions_list in generator:
                # This is for 1 (summary, references) pair
                input_questions_list = []
                input_answers_list = []
                input_predictions_list = []
                input_probabilities_list = []
                input_null_probabilities_list = []
                for qa_pairs, predictions in zip(qa_pairs_list, predictions_list):
                    # This is the set of QA pairs for 1 reference
                    input_questions_list.append([])
                    input_answers_list.append([])
                    input_predictions_list.append([])
                    input_probabilities_list.append([])
                    input_null_probabilities_list.append([])
                    for qa, prediction in zip(qa_pairs, predictions):
                        input_questions_list[-1].append(qa['question'])
                        input_answers_list[-1].append(qa['answer'])
                        input_predictions_list[-1].append(prediction['prediction'])
                        input_probabilities_list[-1].append(prediction['probability'])
                        input_null_probabilities_list[-1].append(prediction['null_probability'])

                metrics, scores = self.scorer.score_multi_ref(
                    summary,
                    input_questions_list,
                    input_answers_list,
                    input_predictions_list,
                    input_probabilities_list,
                    input_null_probabilities_list
                )
                metrics = MetricsDict({'qa-eval': metrics})
                metrics_list.append(metrics)
                scores_list.append(scores)

            logger.info('Finished scoring predictions')
            return metrics_list, scores_list

        def _combine_outputs(self,
                             metrics_list: List[MetricsDict],
                             qa_pairs_lists: List[List[List[Dict[str, Any]]]],
                             predictions_lists: List[List[List[Dict[str, Any]]]],
                             scores_lists: List[List[List[Dict[str, float]]]]) -> List[List[List[Dict[str, Any]]]]:
            # This method will combine the metrics and QA pair metadata together into a tuple so they can
            # both be returned together
            combined = []
            for metrics, qa_pairs_list, predictions_list, scores_list in zip(metrics_list, qa_pairs_lists, predictions_lists, scores_lists):
                # This is for 1 (summary, reference) pair
                combined.append((metrics, []))
                for qa_pairs, predictions, scores in zip(qa_pairs_list, predictions_list, scores_list):
                    # This is for 1 reference
                    combined[-1][1].append([])
                    for qa, prediction, score in zip(qa_pairs, predictions, scores):
                        prediction = dict(**prediction)
                        for key in self.scorer.keys():
                            prediction[key] = score[key]
                        combined[-1][1][-1].append({'question': qa, 'prediction': prediction})
            return combined

        def _insert_empty_outputs(self,
                                  metrics_list: List[MetricsDict],
                                  is_empty_list: List[bool],
                                  include_qa_list: bool) -> List[Any]:
            full_metrics_list = []
            index = 0
            for is_empty in is_empty_list:
                if is_empty:
                    empty_metrics = MetricsDict({'qa-eval': self.scorer.default_scores()})
                    if include_qa_list:
                        full_metrics_list.append((empty_metrics, []))
                    else:
                        full_metrics_list.append(empty_metrics)
                else:
                    full_metrics_list.append(metrics_list[index])
                    index += 1
            return full_metrics_list

        def _realign_output(self,
                            summaries_list: List[List[SummaryType]],
                            output: List[Any]) -> List[List[Any]]:
            realigned = []
            index = 0
            for summaries in summaries_list:
                realigned.append([])
                for _ in summaries:
                    realigned[-1].append(output[index])
                    index += 1
            return realigned

        def score_multi_all(self,
                            summaries_list: List[List[SummaryType]],
                            references_list: List[List[ReferenceType]],
                            return_qa_pairs: bool = False) -> List[List[MetricsDict]]:
            summaries_list = self._flatten_summaries(summaries_list)
            references_list = self._flatten_summaries(references_list)

            # Unroll the summaries into lists of parallel (summary, references) pairs. The additional list around
            # the summaries can potentially make the code confusing. Unrolling the list makes it slightly easier to understand
            unrolled_summaries, unrolled_references_list = self._unroll_summaries(summaries_list, references_list)

            # Remove any input summaries that are empty. They mess up the processing otherwise
            unrolled_summaries, unrolled_references_list, is_empty_list = self._get_empty_summary_mask(unrolled_summaries, unrolled_references_list)

            qa_pairs_lists = self._generate_qa_pairs(unrolled_references_list)
            predictions_lists = self._answer_questions(unrolled_summaries, qa_pairs_lists)
            metrics_list, scores_lists = self._score_predictions(unrolled_summaries, qa_pairs_lists, predictions_lists)

            if return_qa_pairs:
                unrolled_output = self._combine_outputs(metrics_list, qa_pairs_lists, predictions_lists, scores_lists)
            else:
                unrolled_output = metrics_list
            unrolled_output = self._insert_empty_outputs(unrolled_output, is_empty_list, return_qa_pairs)

            output = self._realign_output(summaries_list, unrolled_output)
            return output


@MetricSetupSubcommand.register('qa-eval')
class QAEvalSetupSubcommand(MetricSetupSubcommand):
    @overrides
    def add_subparser(self, parser: argparse._SubParsersAction):
        description = 'Setup the QAEval metric'
        self.parser = parser.add_parser('qa-eval', description=description, help=description)
        self.parser.add_argument('--force', action='store_true', help='Forces redownloading the models')
        self.parser.set_defaults(subfunc=self.run)

    @overrides
    def run(self, args):
        print('This setup command will download the necessary model files. It will not install "qaeval". You must "pip install qaeval" on your own.')

        generation_model_id = '1vVhRgLtsQDAOmxYhY5PMPnxxHUyCOdQU'
        generation_model_path = f'{DATA_ROOT}/metrics/qaeval/models/generation/model.tar.gz'
        if args.force and os.path.exists(generation_model_path):
            os.remove(generation_model_path)
        if not os.path.exists(generation_model_path):
            download_file_from_google_drive(generation_model_id, generation_model_path)
        else:
            print('Skipping downloading generation model')

        answering_model_id = '1q2Z3FPP9AYNz0RJKHMlaweNhmLQoyPA8'
        answering_model_zip_path = f'{DATA_ROOT}/metrics/qaeval/models/answering/model.zip'
        answering_model_path = f'{DATA_ROOT}/metrics/qaeval/models/answering/model'
        if args.force:
            if os.path.exists(answering_model_zip_path):
                os.remove(answering_model_zip_path)
            if os.path.exists(answering_model_path):
                shutil.rmtree(answering_model_path)

        if not os.path.exists(answering_model_zip_path):
            download_file_from_google_drive(answering_model_id, answering_model_zip_path)
        else:
            print('Skipping downloading answering model')
        if not os.path.exists(answering_model_path):
            print('Unzipping answering model')
            with zipfile.ZipFile(answering_model_zip_path) as zip:
                zip.extractall(answering_model_path)
        if os.path.exists(answering_model_zip_path):
            os.remove(answering_model_zip_path)

        lerc_model_id = '193K7v6pjOtuXdlMenQW-RzF6ft-xY2qd'
        lerc_model_path = f'{DATA_ROOT}/metrics/qaeval/models/lerc/model.tar.gz'
        if args.force and os.path.exists(lerc_model_path):
            os.remove(lerc_model_path)
        if not os.path.exists(lerc_model_path):
            download_file_from_google_drive(lerc_model_id, lerc_model_path)
        else:
            print('Skipping downloading LERC model')

        lerc_pretrained_model_id = '1fWBahDT-O1mpsbND300cuZuF73mfObzH'
        lerc_pretrained_model_path = f'{DATA_ROOT}/metrics/qaeval/models/lerc/pretrained.tar.gz'
        if args.force and os.path.exists(lerc_pretrained_model_path):
            os.remove(lerc_pretrained_model_path)
        if not os.path.exists(lerc_pretrained_model_path):
            download_file_from_google_drive(lerc_pretrained_model_id, lerc_pretrained_model_path)
        else:
            print('Skipping downloading LERC pretrained model')

        print('Downloading models complete')



import bz2
import gzip
import jsons
import os
from typing import Any


class JsonlWriter(object):
    """
    The ``JsonlWriter`` is a layer of abstraction around writing data to jsonl
    files. The writer will automatically serialize the input objects into json
    strings, then write them to an output file, one object per line. The data
    will be written as plain text or as bytes, depending on the extension of
    the output file. Current supported binary formats are gzip (``.gz``) and
    bz2 (``.bz2``). All other extensions will use plain text.

    The class should be used the same way that a built-in file handler works::

        with JsonlWriter('/path/to/file.jsonl.gz') as out:
            data = ...  # some data to serialize
            out.write(data)

    Parameters
    ----------
    file_path: ``str``
        The path to the file where the data should be written.
    """
    def __init__(self, file_path: str) -> None:
        self.file_path = file_path

    def __enter__(self):
        dirname = os.path.dirname(self.file_path)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        if self.file_path.endswith('.gz'):
            self.file_handler = gzip.open(self.file_path, 'wb')
            self.binary = True
        elif self.file_path.endswith('.bz2'):
            self.file_handler = bz2.open(self.file_path, 'wb')
            self.binary = True
        else:
            self.file_handler = open(self.file_path, 'w')
            self.binary = False
        return self

    def write(self, object: Any) -> None:
        """
        Serializes the input object to a json string and writes it to the file.

        Parameters
        ----------
        object: ``Any``
            The object to write to the file.
        """
        string = jsons.dumps(object)
        if self.binary:
            self.file_handler.write(string.encode() + b'\n')
        else:
            self.file_handler.write(string + '\n')

    def __exit__(self, *args):
        self.file_handler.close()



import tarfile
from io import BytesIO
from lxml import etree
from tqdm import tqdm
from typing import Dict, List

from sacrerouge.common.util import download_url_to_file
from sacrerouge.io import JsonlWriter


def _extract_data(xml_bytes: bytes) -> Dict:
    root = etree.parse(BytesIO(xml_bytes))

    try:
        doc_id = root.xpath('./head/docdata/doc-id')[0].get('id-string')

        year = root.xpath('./head/meta[@name="publication_year"]')[0].get('content')
        month = root.xpath('./head/meta[@name="publication_month"]')[0].get('content')
        day = root.xpath('./head/meta[@name="publication_day_of_month"]')[0].get('content')

        # We prefer hl1, but fallback to hl2 if it does not exist
        nodes = root.xpath('./body/body.head/hedline/hl1')
        if len(nodes) == 0:
            nodes = root.xpath('./body/body.head/hedline/hl2')
        headline = nodes[0].text.strip()
        assert len(headline) > 0

        abstract = []
        for node in root.xpath('./body/body.head/abstract/p'):
            text = node.text.strip()
            if text:
                abstract.append(text)
        assert len(abstract) == 1
        abstract = abstract[0]

        document = []
        for node in root.xpath('./body/body.content/block[@class="full_text"]/p'):
            text = node.text.strip()
            if text:
                document.append(text)
        assert len(document) > 0

        return {
            'instance_id': doc_id,
            'document': {
                'date': f'{year}-{month}-{day}',
                'headline': headline,
                'text': document,
            },
            'summary': {
                'text': abstract
            }
        }
    except:
        print(xml_bytes.decode())


def _save(instances: List, file_path: str) -> None:
    with JsonlWriter(file_path) as out:
        for instance in instances:
            out.write(instance)


def setup(ldc2008t19_tgz: str, output_dir: str, force: bool) -> None:
    # Download the splits from https://github.com/jiacheng-xu/DiscoBERT/tree/release/data_preparation/urls_nyt
    for split in ['train', 'valid', 'test']:
        download_url_to_file(f'https://github.com/jiacheng-xu/DiscoBERT/raw/release/data_preparation/urls_nyt/mapping_{split}.txt',
                             f'{output_dir}/raw/mapping_{split}.txt',
                             force=force)

    train_ids = set(map(int, open(f'{output_dir}/raw/mapping_train.txt', 'r').read().splitlines()))
    valid_ids = set(map(int, open(f'{output_dir}/raw/mapping_valid.txt', 'r').read().splitlines()))
    test_ids = set(map(int, open(f'{output_dir}/raw/mapping_test.txt', 'r').read().splitlines()))
    all_ids = train_ids | valid_ids | test_ids

    train, valid, test = [], [], []
    with tarfile.open(ldc2008t19_tgz, 'r') as tar:
        # Obtain the members to make the tqdm output make sense
        members = []
        for member in tar.getmembers():
            if not member.name.startswith('nyt_corpus/data') or not member.name.endswith('.tgz'):
                continue
            members.append(member)

        for member in tqdm(members):
            tar_bytes = tar.extractfile(member).read()
            with tarfile.open(fileobj=BytesIO(tar_bytes)) as inner_tar:
                for inner_member in inner_tar.getmembers():
                    if not inner_member.name.endswith('.xml'):
                        continue

                    # "01/26/1459527.xml"
                    file_id = int(inner_member.name.split('/')[-1][:-4])
                    if file_id in all_ids:
                        xml_bytes = inner_tar.extractfile(inner_member).read()
                        instance = _extract_data(xml_bytes)

                        if file_id in train_ids:
                            train.append(instance)
                        elif file_id in valid_ids:
                            valid.append(instance)
                        elif file_id in test_ids:
                            test.append(instance)

    assert len(train) == 137778, f'Train has {len(train)} instances, expected 137778'
    assert len(valid) == 17222, f'Valid has {len(valid)} instances, expected 17222'
    assert len(test) == 17223, f'Test has {len(test)} instances, expected 17223'

    _save(train, f'{output_dir}/train.jsonl.gz')
    _save(valid, f'{output_dir}/valid.jsonl.gz')
    _save(test, f'{output_dir}/test.jsonl.gz')


import argparse
import unittest
from typing import List, Tuple

from sacrerouge.common.arguments import add_metric_arguments, get_metric_from_arguments, get_dataset_reader_from_argument
from sacrerouge.data.dataset_readers import DatasetReader
from sacrerouge.metrics import Metric


@DatasetReader.register('faked')
class _FakedDatasetReader(DatasetReader):
    def __init__(self, x: int = 4):
        super().__init__()
        self.x = x


@Metric.register('faked')
class _FakedMetric(Metric):
    def __init__(self,
                 a,
                 b: bool,
                 c: int,
                 d: float = 3.5,
                 e: str = None,
                 f: bool = False,
                 g: bool = True,
                 h: List[str] = None,
                 i: List[str] = list(['a', 'b'])):
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.e = e
        self.f = f
        self.g = g
        self.h = h
        self.i = i


class TestArguments(unittest.TestCase):
    def test_get_dataset_reader_from_argument(self):
        reader = get_dataset_reader_from_argument('faked')
        assert isinstance(reader, _FakedDatasetReader)
        assert reader.x == 4

        reader = get_dataset_reader_from_argument('{"type": "faked", "x": 3}')
        assert isinstance(reader, _FakedDatasetReader)
        assert reader.x == 3

    def test_add_metric_arguments(self):
        argp = argparse.ArgumentParser()
        add_metric_arguments(argp, _FakedMetric)

        dest_to_action = {action.dest: action for action in argp._actions}
        assert len(dest_to_action) == 10  # Extra argument is for "help"

        action = dest_to_action['a']
        assert action.type == str
        assert action.default is None
        assert action.required is True

        # Boolean type is a function which is not checked here
        action = dest_to_action['b']
        assert action.default is None
        assert action.required is True

        action = dest_to_action['c']
        assert action.type == int
        assert action.default is None
        assert action.required is True

        action = dest_to_action['d']
        assert action.type == float
        assert action.default == 3.5
        assert action.required is False

        action = dest_to_action['e']
        assert action.type == str
        assert action.default == None
        assert action.required is False

        # Boolean type is a function which is not checked here
        action = dest_to_action['f']
        assert action.default == False
        assert action.required is False

        # Boolean type is a function which is not checked here
        action = dest_to_action['g']
        assert action.default is True
        assert action.required is False

        action = dest_to_action['h']
        assert action.type == str
        assert action.default == None
        assert action.required is False

        action = dest_to_action['i']
        assert action.type == str
        assert action.default == '["a", "b"]'
        assert action.required is False

    def test_get_metric_from_arguments(self):
        argp = argparse.ArgumentParser()
        add_metric_arguments(argp, _FakedMetric)

        # All arguments provided and defaults overridden
        args = argp.parse_args([
            '--a', 'test',
            '--b', 'true',
            '--c', '4',
            '--d', '1.2',
            '--e', 'hello',
            '--f', 'true',
            '--g', 'false',
            '--h', '["1", "2"]',
            '--i', '["3", "4"]'
        ])
        metric = get_metric_from_arguments(_FakedMetric, args)
        assert isinstance(metric, _FakedMetric)
        assert metric.a == 'test'
        assert metric.b is True
        assert metric.c == 4
        assert metric.d == 1.2
        assert metric.e == 'hello'
        assert metric.f is True
        assert metric.g is False
        assert metric.h == ['1', '2']
        assert metric.i == ['3', '4']

        # Only required arguments
        args = argp.parse_args([
            '--a', 'test',
            '--b', 'true',
            '--c', '4'
        ])
        metric = get_metric_from_arguments(_FakedMetric, args)
        assert isinstance(metric, _FakedMetric)
        assert metric.a == 'test'
        assert metric.b is True
        assert metric.c == 4
        assert metric.d == 3.5
        assert metric.e is None
        assert metric.f is False
        assert metric.g is True
        assert metric.h is None
        assert metric.i == ['a', 'b']


import bz2
import gzip
import jsons
from typing import Any, List, Optional, Type

from sacrerouge.io.util import is_gz_file


class JsonlReader(object):
    """
    The `JsonlReader` is a layer of abstraction around reading serialized
    objects from a jsonl file. The reader will automatically deserialize and return
    one object from each line in the file. The data in the file will be decoded
    from a binary file depending on the extension of the file name. Current
    supported binary formats are gzip (``.gz``) and bz2 (``.bz2``). For gzip only,
    this will also inspect the file to see if it's gzipped in addition to checking
    the extension.

    The class should be used the same way that a built-in file handler works::

        with JsonlReader('/path/to/file.jsonl.gz') as f:
            for data in f:
                ...

    The class uses the cached path functionality from AllenNLP, so it is also
    possible to pass a url to the constructor.

    Parameters
    ----------
    file_path: ``str``
        The path to the file where the data should be read.
    cls: ``Optional[Type]``
        The type that the object should be loaded as.
    """
    def __init__(self, file_path: str, cls: Optional[Type] = None) -> None:
        self.file_path = file_path
        self.cls = cls

    def __enter__(self):
        self.binary = False
        if self.file_path.endswith('.gz') or is_gz_file(self.file_path):
            self.file_handler = gzip.open(self.file_path, 'rb')
            self.binary = True
        elif self.file_path.endswith('.bz2'):
            self.file_handler = bz2.open(self.file_path, 'rb')
            self.binary = True
        else:
            self.file_handler = open(self.file_path, 'r')
            self.binary = False
        return self

    def __iter__(self):
        return self

    def __next__(self) -> Any:
        for line in self.file_handler:
            if self.binary:
                line = line.decode()
            return jsons.loads(line, self.cls)
        raise StopIteration

    def __exit__(self, *args):
        self.file_handler.close()

    def read(self) -> List[Any]:
        """Reads all of the instances into a list."""
        with self:
            return [instance for instance in self]



import argparse
import logging
import os
import sys
from collections import defaultdict
from overrides import overrides
from subprocess import Popen, PIPE
from typing import List, Optional, Tuple

from sacrerouge.commands import MetricSetupSubcommand
from sacrerouge.common import DATA_ROOT, TemporaryDirectory
from sacrerouge.common.util import download_file_from_google_drive
from sacrerouge.data import MetricsDict
from sacrerouge.data.types import ReferenceType
from sacrerouge.data.types import SummaryType
from sacrerouge.metrics import Metric, ReferenceBasedMetric

logger = logging.getLogger(__name__)


@Metric.register('rouge')
class Rouge(ReferenceBasedMetric):
    def __init__(self,
                 max_ngram: int = 4,
                 use_porter_stemmer: bool = True,
                 remove_stopwords: bool = False,
                 max_bytes: Optional[int] = None,
                 max_words: Optional[int] = None,
                 compute_rouge_l: bool = False,
                 skip_bigram_gap_length: Optional[int] = None,
                 wlcs_weight: Optional[float] = None,
                 rouge_root: str = f'{DATA_ROOT}/metrics/ROUGE-1.5.5',
                 scoring_function: str = 'average'):
        super().__init__()
        self.max_ngram = max_ngram
        self.use_porter_stemmer = use_porter_stemmer
        self.remove_stopwords = remove_stopwords
        self.max_bytes = max_bytes
        self.max_words = max_words
        self.compute_rouge_l = compute_rouge_l
        self.skip_bigram_gap_length = skip_bigram_gap_length
        self.wlcs_weight = wlcs_weight
        self.rouge_script_location = f'{rouge_root}/ROUGE-1.5.5.pl'
        self.rouge_eval_home = f'{rouge_root}/data'
        self.scoring_function = scoring_function

        if not os.path.exists(rouge_root):
            raise Exception(f'Path "{rouge_root}" does not exist. Have you setup ROUGE?')

    def _save_summary(self, summary: SummaryType, file_path: str) -> None:
        dirname = os.path.dirname(file_path)
        os.makedirs(dirname, exist_ok=True)
        with open(file_path, 'w') as out:
            if isinstance(summary, list):
                for sentence in summary:
                    out.write(sentence + '\n')
            else:
                out.write(summary)

    def _save_config_file(self,
                          config_filename: str,
                          summary_filenames_list: List[List[str]],
                          reference_filenames_list: List[List[str]]):
        output_dir = os.path.dirname(config_filename)
        with open(config_filename, 'w') as out:
            out.write(f'<ROUGE_EVAL version="1.0">\n')
            for i, (reference_filenames, summary_filenames) in enumerate(zip(reference_filenames_list, summary_filenames_list)):
                out.write(f'<EVAL ID="{i + 1}">\n')
                out.write(f'<INPUT-FORMAT TYPE="SPL"></INPUT-FORMAT>\n')
                out.write(f'<PEER-ROOT>{output_dir}</PEER-ROOT>\n')
                out.write(f'<MODEL-ROOT>{output_dir}</MODEL-ROOT>\n')
                out.write(f'<PEERS>\n')
                for j, summary_filename in enumerate(summary_filenames):
                    out.write(f'<P ID="{j + 1}">{summary_filename}</P>\n')
                out.write(f'</PEERS>\n')
                out.write(f'<MODELS>\n')
                for j, reference_filename in enumerate(reference_filenames):
                    symbol = chr(j + 65)
                    out.write(f'<M ID="{symbol}">{reference_filename}</M>\n')
                out.write(f'</MODELS>\n')
                out.write(f'</EVAL>\n')
            out.write(f'</ROUGE_EVAL>\n')

    def _parse_average_line(self, columns: List[str]) -> Tuple[str, str, float, float, float]:
        assert len(columns) == 8
        if columns[2][-2] == 'R':
            metric = 'recall'
        elif columns[2][-2] == 'P':
            metric = 'precision'
        elif columns[2][-2] == 'F':
            metric = 'f1'
        else:
            raise Exception(f'Unknown metric: {columns[2]}')

        summarizer_id = int(columns[0]) - 1
        value = float(columns[3]) * 100
        lower_ci = float(columns[5]) * 100
        upper_ci = float(columns[7][:-1]) * 100
        return summarizer_id, metric, value, lower_ci, upper_ci

    def _parse_individual_line(self, columns: List[str]) -> Tuple[int, int, float, float, float]:
        assert len(columns) == 7
        period = columns[3].index('.')
        instance_id = int(columns[3][:period]) - 1
        summarizer_id = int(columns[3][period + 1:]) - 1
        recall = float(columns[4][2:]) * 100
        precision = float(columns[5][2:]) * 100
        f1 = float(columns[6][2:]) * 100
        return instance_id, summarizer_id, recall, precision, f1

    def _parse_rouge_stdout(self, stdout: str):
        lines = stdout.splitlines()
        macro_metrics_dict = defaultdict(lambda: defaultdict(MetricsDict))
        micro_metrics_dicts = defaultdict(lambda: defaultdict(MetricsDict))
        for line in lines:
            if line in ['---------------------------------------------', '.............................................']:
                continue
            columns = line.split()
            rouge_metric = columns[1].lower()
            if columns[2] == 'Eval':
                instance_id, summarizer_id, recall, precision, f1 = self._parse_individual_line(columns)
                micro_metrics_dicts[instance_id][summarizer_id][rouge_metric] = {
                    'recall': recall,
                    'precision': precision,
                    'f1': f1
                }
            else:
                summarizer_id, metric, value, lower_ci, upper_ci = self._parse_average_line(columns)
                macro_metrics_dict[summarizer_id][rouge_metric][metric] = value

        # Flatten the metrics into arrays instead of dicts
        macro_metrics_list = [None] * len(macro_metrics_dict)
        for summarizer_id, metrics in macro_metrics_dict.items():
            macro_metrics_list[summarizer_id] = metrics

        micro_metrics_lists = [None] * len(micro_metrics_dicts)
        for instance_id, metrics_dict in micro_metrics_dicts.items():
            micro_metrics_lists[instance_id] = [None] * len(metrics_dict)
            for summarizer_id, metrics in metrics_dict.items():
                micro_metrics_lists[instance_id][summarizer_id] = metrics
        return macro_metrics_list, micro_metrics_lists

    def _run(self,
             summaries_list: List[List[SummaryType]],
             references_list: List[List[SummaryType]]) -> Tuple[List[MetricsDict], List[List[MetricsDict]]]:
        with TemporaryDirectory() as temp_dir:
            summary_filenames_list = []
            reference_filenames_list = []

            for i, (summaries, references) in enumerate(zip(summaries_list, references_list)):
                summary_filenames_list.append([])
                reference_filenames_list.append([])
                for j, summary in enumerate(summaries):
                    summary_filename = f'{i}/model.{j}.txt'
                    summary_filenames_list[-1].append(summary_filename)
                    self._save_summary(summary, f'{temp_dir}/{summary_filename}')

                for j, reference in enumerate(references):
                    symbol = chr(j + 65)
                    reference_filename = f'{i}/gold.{symbol}.txt'
                    reference_filenames_list[-1].append(reference_filename)
                    self._save_summary(reference, f'{temp_dir}/{reference_filename}')

            config_filename = f'{temp_dir}/config.xml'
            self._save_config_file(config_filename, summary_filenames_list, reference_filenames_list)

            command = [
                self.rouge_script_location,
                '-e', self.rouge_eval_home,
                '-n', str(self.max_ngram),
                '-a',
                '-c', '95',
                '-r', '1000',
                '-p', '0.5',
                '-t', '0',
                '-d'
            ]
            if self.use_porter_stemmer:
                command += ['-m']
            if self.remove_stopwords:
                command += ['-s']
            if self.max_bytes is not None:
                command += ['-b', str(self.max_bytes)]
            if self.max_words is not None:
                command += ['-l', str(self.max_words)]
            if not self.compute_rouge_l:
                command += ['-x']
            if self.skip_bigram_gap_length is not None:
                command += ['-2', str(self.skip_bigram_gap_length), '-u']
            if self.wlcs_weight is not None:
                command += ['-w', str(self.wlcs_weight)]
            if self.scoring_function == 'average':
                command += ['-f', 'A']
            elif self.scoring_function == 'max':
                command += ['-f', 'B']
            else:
                raise Exception(f'Unrecognized scoring function: "{self.scoring_function}"')
            command += [config_filename]

            # We used to fail if anything was written to stderr, but ROUGE writes
            # a warning if the number of peers per reference set is different, which
            # is expected in some situations for us (if we just have more summaries
            # to score for some reference sets than others). Therefore, we no longer fail
            # if stderr is not empty.
            logger.info(f'Running ROUGE command: "{" ".join(command)}"')
            process = Popen(command, stdout=PIPE, stderr=PIPE)
            stdout, stderr = process.communicate()

            macro_metrics_list, micro_metrics_lists = self._parse_rouge_stdout(stdout.decode())
            return macro_metrics_list, micro_metrics_lists

    def score_multi_all(self,
                        summaries_list: List[List[SummaryType]],
                        references_list: List[List[ReferenceType]]) -> List[List[MetricsDict]]:
        _, micro_metrics_lists = self._run(summaries_list, references_list)
        return micro_metrics_lists

    def evaluate(self,
                 summaries: List[SummaryType],
                 references_list: List[List[ReferenceType]]) -> Tuple[MetricsDict, List[MetricsDict]]:
        summaries_list = [[summary] for summary in summaries]
        macro_metrics_list, micro_metrics_lists = self._run(summaries_list, references_list)

        macro_metrics = macro_metrics_list[0]
        micro_metrics_list = [metrics_list[0] for metrics_list in micro_metrics_lists]
        return macro_metrics, micro_metrics_list


@MetricSetupSubcommand.register('rouge')
class RougeSetupSubcommand(MetricSetupSubcommand):
    @overrides
    def add_subparser(self, parser: argparse._SubParsersAction):
        description = 'Setup the ROUGE metric'
        self.parser = parser.add_parser('rouge', description=description, help=description)
        self.parser.set_defaults(subfunc=self.run)

    @overrides
    def run(self, args):
        print(f'Downloading ROUGE-1.5.5')

        # I believe the ROUGE data files are platform-dependent. We verify it runs later, but in case it is, we have
        # to download the data files for the platform.
        # https://stackoverflow.com/questions/8220108/how-do-i-check-the-operating-system-in-python
        if sys.platform in ['linux', 'linux2']:
            file_id = '1K4J2wHGjAyr3LoSgaQuWZ_YyjtUGf26m'
        elif sys.platform == 'darwin':
            file_id = '1y0rDnTplQ83b2PQu_TgezbFpGOthP0gG'
        else:
            # No idea -- default to Linux
            file_id = '1K4J2wHGjAyr3LoSgaQuWZ_YyjtUGf26m'

        download_file_from_google_drive(file_id, f'{DATA_ROOT}/metrics/ROUGE-1.5.5.zip')

        commands = [
            f'cd {DATA_ROOT}/metrics',
            f'unzip ROUGE-1.5.5.zip',
            f'rm ROUGE-1.5.5.zip'
        ]
        command = ' && '.join(commands)

        process = Popen(command, shell=True)
        process.communicate()
        if process.returncode != 0:
            print('ROUGE setup failure')

        # ROUGE has data files which may not successfully load. Therefore, if it fails to run on a simple example,
        # the user needs to run some perl code within the ROUGE directory to correct the data file
        try:
            summary = 'Dan walked to the bakery this morning.'
            reference = 'Dan went to buy scones earlier this morning.'
            rouge = Rouge()
            rouge.score(summary, [reference])
            print('ROUGE setup success')
        except IndexError:
            print('ROUGE setup failure')
            print('It is very likely that either (1) you need to install the Perl XML::DOM library or '
                  '(2) you need to rebuild the ROUGE database file. The first case is more likely if you '
                  'are running on Linux or MacOS. See '
                  'https://github.com/danieldeutsch/sacrerouge/blob/master/doc/metrics/rouge.md for '
                  'instructions for each of these steps. Afterward, this example should run without failing:')
            print()
            print('>>> from sacrerouge.metrics import Rouge')
            print('>>> ')
            print('>>> summary = "Dan walked to the bakery this morning."')
            print('>>> reference = "Dan went to buy scones earlier this morning."')
            print('>>> ')
            print('>>> rouge = Rouge()')
            print('>>> rouge.score(summary, [reference])')



import pytest

from sacrerouge.common.testing.metric_test_cases import ReferenceBasedMetricTestCase
from sacrerouge.common.testing.util import sacrerouge_command_exists
from sacrerouge.metrics import MoverScore
from sacrerouge.metrics.moverscore import MOVERSCORE_INSTALLED


@pytest.mark.skipif(not MOVERSCORE_INSTALLED, reason='MoverScore not setup')
class TestMoverScore(ReferenceBasedMetricTestCase):
    def test_moverscore(self):
        # This is a regression test, not necessarily a test for correctness
        metric = MoverScore()
        expected_output = [
            {'MoverScore': 0.13538047261472216},
            {'MoverScore': 0.19151070384024393},
            {'MoverScore': 0.21223604036747462},
            {'MoverScore': 0.17447731600410432},
            {'MoverScore': 0.1479818362434758},
            {'MoverScore': 0.14823672247671074},
            {'MoverScore': 0.22382319193839145},
            {'MoverScore': 0.23362081109521693},
            {'MoverScore': 0.20162050907949902},
            {'MoverScore': 0.24070720220937813},
            {'MoverScore': 0.22734330926367746},
            {'MoverScore': 0.18901058452094593}
        ]
        super().assert_expected_output(metric, expected_output)

    def test_moverscore_order_invariant(self):
        metric = MoverScore()
        self.assert_order_invariant(metric)

    def test_command_exists(self):
        assert sacrerouge_command_exists(['moverscore'])

    def test_setup_command_exists(self):
        assert sacrerouge_command_exists(['setup-metric', 'moverscore'])


import functools
import numpy as np
import unittest
from scipy.stats import kendalltau, pearsonr, spearmanr

from sacrerouge.data import Metrics
from sacrerouge.stats import convert_to_matrices, summary_level_corr, system_level_corr, global_corr, \
    bootstrap_system_sample, bootstrap_input_sample, bootstrap_both_sample, bootstrap_ci, fisher_ci, corr_ci, \
    random_bool_mask, permute_systems, permute_inputs, permute_both, bootstrap_diff_test, permutation_diff_test, \
    williams_diff_test, corr_diff_test, bonferroni_partial_conjunction_pvalue_test


class TestStats(unittest.TestCase):
    def test_convert_to_matrices(self):
        metrics_list = [
            Metrics('1', 'A', 'peer', {'m1': 1, 'm2': 2, 'm3': 3}),
            Metrics('2', 'A', 'peer', {'m1': 4, 'm2': 5}),
            Metrics('1', 'B', 'peer', {'m1': 6, 'm2': 7, 'm3': 8}),
            Metrics('2', 'B', 'peer', {'m1': 9, 'm2': 10, 'm3': 11}),
        ]
        m1 = convert_to_matrices(metrics_list, 'm1')
        np.testing.assert_array_equal(m1, [[1, 4], [6, 9]])

        m1, m2 = convert_to_matrices(metrics_list, 'm1', 'm2')
        np.testing.assert_array_equal(m1, [[1, 4], [6, 9]])
        np.testing.assert_array_equal(m2, [[2, 5], [7, 10]])

        m3 = convert_to_matrices(metrics_list, 'm3')
        np.testing.assert_array_equal(m3, [[3, np.nan], [8, 11]])

        metrics_list = [
            Metrics('1', 'A', 'peer', {'m1': 1, 'm2': 2}),
            Metrics('2', 'A', 'peer', {'m1': 4, 'm2': 5}),
            Metrics('1', 'B', 'peer', {'m1': 6, 'm2': 7}),
            Metrics('3', 'B', 'peer', {'m1': 2, 'm2': 9}),
        ]
        m1 = convert_to_matrices(metrics_list, 'm1')
        np.testing.assert_array_equal(m1, [[1, 4, np.nan], [6, np.nan, 2]])

    def test_summary_level_corr(self):
        # This will end up skipping the last column because the scores are identical,
        # so the correlation is NaN
        X = np.array([
            [1, 9, 2],
            [4, 5, 2],
            [6, 7, 2]
        ])
        Y = np.array([
            [11, 12, 13],
            [14, 15, 16],
            [17, 18, 19]
        ])
        r = summary_level_corr(pearsonr, X, Y)
        self.assertAlmostEqual(r, 0.2466996339, places=4)

        X = np.array([
            [1, 2],
            [1, 2],
        ])
        Y = np.array([
            [11, 12],
            [14, 15],
        ])
        # This shouldn't have any correlations because both are NaN
        assert summary_level_corr(pearsonr, X, Y) is None

        X = np.array([
            [1, 9, 2],
            [np.nan, 5, 4],
            [6, 7, 7]
        ])
        Y = np.array([
            [11, 12, 13],
            [np.nan, 15, 16],
            [17, 18, 19]
        ])
        self.assertAlmostEqual(summary_level_corr(pearsonr, X, Y), 0.4977997559)

        # Fails because they do not have parallel nans
        X = np.array([
            [1, 9, 2],
            [4, np.nan, 2],
            [6, 7, 2]
        ])
        Y = np.array([
            [11, 12, np.nan],
            [14, 15, 16],
            [17, 18, 19]
        ])
        with self.assertRaises(Exception):
            summary_level_corr(pearsonr, X, Y)

    def test_system_level_corr(self):
        X = np.array([
            [1, 9, 2],
            [4, 5, 2],
            [6, 7, 2]
        ])
        Y = np.array([
            [11, 12, 13],
            [14, 15, 16],
            [17, 18, 19]
        ])
        r = system_level_corr(pearsonr, X, Y)
        self.assertAlmostEqual(r, 0.7205766921, places=4)

        r, pvalue = system_level_corr(pearsonr, X, Y, return_pvalue=True)
        self.assertAlmostEqual(r, 0.7205766921, places=4)
        self.assertAlmostEqual(pvalue, 0.48775429164459994, places=4)

        X = np.array([
            [1, 9, 2],
            [4, 5, np.nan],
            [6, np.nan, 2]
        ])
        Y = np.array([
            [11, 12, 13],
            [14, 15, np.nan],
            [17, np.nan, 19]
        ])
        r = system_level_corr(pearsonr, X, Y)
        self.assertAlmostEqual(r, -0.09578262852, places=4)

        r, pvalue = system_level_corr(pearsonr, X, Y, return_pvalue=True)
        self.assertAlmostEqual(r, -0.09578262852, places=4)
        self.assertAlmostEqual(pvalue, 0.938929260614949, places=4)

        X = np.array([
            [1, 2],
            [1, 2],
        ])
        Y = np.array([
            [11, 12],
            [14, 15],
        ])
        # This shouldn't have any correlations because the average of X is all the same
        assert system_level_corr(pearsonr, X, Y) is None
        assert system_level_corr(pearsonr, X, Y, return_pvalue=True) == (None, None)

        # Fails because they do not have parallel nans
        X = np.array([
            [1, 9, 2],
            [4, np.nan, 2],
            [6, 7, 2]
        ])
        Y = np.array([
            [11, 12, np.nan],
            [14, 15, 16],
            [17, 18, 19]
        ])
        with self.assertRaises(Exception):
            system_level_corr(pearsonr, X, Y)

    def test_global_corr(self):
        X = np.array([
            [1, 9, 2],
            [4, 5, 2],
            [6, 7, 2]
        ])
        Y = np.array([
            [11, 12, 13],
            [14, 15, 16],
            [17, 18, 19]
        ])
        r = global_corr(pearsonr, X, Y)
        self.assertAlmostEqual(r, 0.06691496051, places=4)

        r, pvalue = global_corr(pearsonr, X, Y, return_pvalue=True)
        self.assertAlmostEqual(r, 0.06691496051, places=4)
        self.assertAlmostEqual(pvalue, 0.8641895868792804, places=4)

        X = np.array([
            [1, 9, 2],
            [np.nan, 5, 2],
            [6, 7, np.nan]
        ])
        Y = np.array([
            [11, 12, 13],
            [np.nan, 15, 16],
            [17, 18, np.nan]
        ])
        r = global_corr(pearsonr, X, Y)
        self.assertAlmostEqual(r, 0.2897249422, places=4)

        r, pvalue = global_corr(pearsonr, X, Y, return_pvalue=True)
        self.assertAlmostEqual(r, 0.2897249422, places=4)
        self.assertAlmostEqual(pvalue, 0.5285282548518477, places=4)

        X = np.array([
            [1, 1],
            [1, 1],
        ])
        Y = np.array([
            [11, 12],
            [14, 15],
        ])
        # This shouldn't have any correlations because X is identical
        assert global_corr(pearsonr, X, Y) is None
        assert global_corr(pearsonr, X, Y, return_pvalue=True) == (None, None)

        # Fails because they do not have parallel nans
        X = np.array([
            [1, 9, 2],
            [4, np.nan, 2],
            [6, 7, 2]
        ])
        Y = np.array([
            [11, 12, np.nan],
            [14, 15, 16],
            [17, 18, 19]
        ])
        with self.assertRaises(Exception):
            global_corr(pearsonr, X, Y)

    def test_bootstrap_system_sample(self):
        A = np.array([
            [1, 2, 3, 4],
            [5, 6, 7, 8],
            [9, 10, 11, 12]
        ])
        B = np.array([
            [13, 14, 15, 16],
            [17, 18, 19, 20],
            [21, 22, 23, 24]
        ])

        # We check what sample should be taken with this random seed
        np.random.seed(4)
        np.testing.assert_array_equal(np.random.choice(3, 3, replace=True), [2, 2, 1])

        np.random.seed(4)
        A_s = bootstrap_system_sample(A)
        np.testing.assert_array_equal(A_s, [[9, 10, 11, 12], [9, 10, 11, 12], [5, 6, 7, 8]])

        np.random.seed(4)
        A_s, B_s = bootstrap_system_sample(A, B)
        np.testing.assert_array_equal(A_s, [[9, 10, 11, 12], [9, 10, 11, 12], [5, 6, 7, 8]])
        np.testing.assert_array_equal(B_s, [[21, 22, 23, 24], [21, 22, 23, 24], [17, 18, 19, 20]])

    def test_bootstrap_input_sample(self):
        A = np.array([
            [1, 2, 3, 4],
            [5, 6, 7, 8],
            [9, 10, 11, 12]
        ])
        B = np.array([
            [13, 14, 15, 16],
            [17, 18, 19, 20],
            [21, 22, 23, 24]
        ])

        # We check what sample should be taken with this random seed
        np.random.seed(4)
        np.testing.assert_array_equal(np.random.choice(4, 4, replace=True), [2, 2, 3, 1])

        np.random.seed(4)
        A_s = bootstrap_input_sample(A)
        np.testing.assert_array_equal(A_s, [[3, 3, 4, 2], [7, 7, 8, 6], [11, 11, 12, 10]])

        np.random.seed(4)
        A_s, B_s = bootstrap_input_sample(A, B)
        np.testing.assert_array_equal(A_s, [[3, 3, 4, 2], [7, 7, 8, 6], [11, 11, 12, 10]])
        np.testing.assert_array_equal(B_s, [[15, 15, 16, 14], [19, 19, 20, 18], [23, 23, 24, 22]])

    def test_bootstrap_both_sample(self):
        A = np.array([
            [1, 2, 3, 4],
            [5, 6, 7, 8],
            [9, 10, 11, 12]
        ])
        B = np.array([
            [13, 14, 15, 16],
            [17, 18, 19, 20],
            [21, 22, 23, 24]
        ])

        # We check what sample should be taken with this random seed
        np.random.seed(4)
        np.testing.assert_array_equal(np.random.choice(3, 3, replace=True), [2, 2, 1])
        np.testing.assert_array_equal(np.random.choice(4, 4, replace=True), [1, 0, 3, 0])

        np.random.seed(4)
        A_s = bootstrap_both_sample(A)
        np.testing.assert_array_equal(A_s, [[10, 9, 12, 9], [10, 9, 12, 9], [6, 5, 8, 5]])

        np.random.seed(4)
        A_s, B_s = bootstrap_both_sample(A, B)
        np.testing.assert_array_equal(A_s, [[10, 9, 12, 9], [10, 9, 12, 9], [6, 5, 8, 5]])
        np.testing.assert_array_equal(B_s, [[22, 21, 24, 21], [22, 21, 24, 21], [18, 17, 20, 17]])

    def test_bootstrap_ci(self):
        # Regression test
        np.random.seed(3)
        X = np.array([
            [1, 2, 3],
            [4, 5, 6],
            [7, 8, 9]
        ])
        Y = np.array([
            [5, 2, 7],
            [1, 7, 3],
            [4, 2, 2]
        ])
        corr_func = functools.partial(global_corr, pearsonr)

        lower, upper = bootstrap_ci(corr_func, X, Y, bootstrap_system_sample)
        self.assertAlmostEqual(lower, -0.8660254037844388, places=4)
        self.assertAlmostEqual(upper, 0.39735970711951324, places=4)

        lower, upper = bootstrap_ci(corr_func, X, Y, bootstrap_system_sample, alpha=0.1)
        self.assertAlmostEqual(lower, -0.5773502691896258, places=4)
        self.assertAlmostEqual(upper, 0.32732683535398865, places=4)

        lower, upper = bootstrap_ci(corr_func, X, Y, bootstrap_input_sample)
        self.assertAlmostEqual(lower, -0.9449111825230679, places=4)
        self.assertAlmostEqual(upper, 0.0, places=4)

        lower, upper = bootstrap_ci(corr_func, X, Y, bootstrap_both_sample)
        self.assertAlmostEqual(lower, -1.0, places=4)
        self.assertAlmostEqual(upper, 1.0, places=4)

    def test_fisher_ci(self):
        pearson_global = functools.partial(global_corr, pearsonr)
        spearman_global = functools.partial(global_corr, spearmanr)
        kendall_global = functools.partial(global_corr, kendalltau)

        pearson_system = functools.partial(system_level_corr, pearsonr)
        spearman_system = functools.partial(system_level_corr, spearmanr)
        kendall_system = functools.partial(system_level_corr, kendalltau)

        pearson_summary = functools.partial(summary_level_corr, pearsonr)
        spearman_summary = functools.partial(summary_level_corr, spearmanr)
        kendall_summary = functools.partial(summary_level_corr, kendalltau)

        np.random.seed(12)
        X = np.random.rand(5, 7)
        Y = np.random.rand(5, 7)

        self.assertAlmostEqual(fisher_ci(pearson_global, X, Y), (-0.02763744135012373, 0.5818846438651135), places=4)
        self.assertAlmostEqual(fisher_ci(spearman_global, X, Y), (-0.06733469087453943, 0.5640758668009686), places=4)
        self.assertAlmostEqual(fisher_ci(kendall_global, X, Y), (-0.029964677270600665, 0.4098565164085108), places=4)

        self.assertAlmostEqual(fisher_ci(pearson_system, X, Y), (-0.6445648014599665, 0.9644395142168088), places=4)
        self.assertAlmostEqual(fisher_ci(spearman_system, X, Y), (-0.6708734441360908, 0.9756771001362685), places=4)
        self.assertAlmostEqual(fisher_ci(kendall_system, X, Y), (-0.7023910748254728, 0.9377789575997956), places=4)

        self.assertAlmostEqual(fisher_ci(pearson_summary, X, Y), (-0.808376631595968, 0.9287863878043723), places=4)
        self.assertAlmostEqual(fisher_ci(spearman_summary, X, Y), (-0.7262127280589684, 0.9653646507719408), places=4)
        self.assertAlmostEqual(fisher_ci(kendall_summary, X, Y), (-0.684486849088761, 0.9418063314024349), places=4)

    def test_corr_ci(self):
        # Regression test
        np.random.seed(3)
        X = np.array([
            [1, 2, 3],
            [4, 5, 6],
            [7, 8, 9]
        ])
        Y = np.array([
            [5, 2, 7],
            [1, 7, 3],
            [4, 2, 2]
        ])
        corr_func = functools.partial(global_corr, pearsonr)

        # Make sure we get the same result going through bootstrap_ci and corr_ci
        expected_lower, expected_upper = bootstrap_ci(corr_func, X, Y, bootstrap_system_sample)
        lower, upper = corr_ci(corr_func, X, Y, 'bootstrap-system')
        self.assertAlmostEqual(lower, expected_lower, places=4)
        self.assertAlmostEqual(upper, expected_upper, places=4)

        expected_lower, expected_upper = bootstrap_ci(corr_func, X, Y, bootstrap_input_sample)
        lower, upper = corr_ci(corr_func, X, Y, 'bootstrap-input')
        self.assertAlmostEqual(lower, expected_lower, places=4)
        self.assertAlmostEqual(upper, expected_upper, places=4)

        expected_lower, expected_upper = bootstrap_ci(corr_func, X, Y, bootstrap_both_sample)
        lower, upper = corr_ci(corr_func, X, Y, 'bootstrap-both')
        self.assertAlmostEqual(lower, expected_lower, places=4)
        self.assertAlmostEqual(upper, expected_upper, places=4)

        # If we do a single tail, the result should be the same with alpha / 2
        expected_lower, expected_upper = bootstrap_ci(corr_func, X, Y, bootstrap_system_sample)
        lower, upper = corr_ci(corr_func, X, Y, 'bootstrap-system', alpha=0.025, two_tailed=False)
        self.assertAlmostEqual(lower, expected_lower, places=4)
        self.assertAlmostEqual(upper, expected_upper, places=4)

        # None cases
        assert corr_ci(corr_func, X, Y, None) == (None, None)
        assert corr_ci(corr_func, X, Y, 'none') == (None, None)

        with self.assertRaises(Exception):
            corr_ci(corr_func, X, Y, 'does-not-exist')

    def test_random_bool_mask(self):
        np.random.seed(7)
        expected_rand = [
            [0.07630829, 0.77991879, 0.43840923, 0.72346518],
            [0.97798951, 0.53849587, 0.50112046, 0.07205113],
            [0.26843898, 0.4998825,  0.67923,    0.80373904]
        ]
        np.testing.assert_array_almost_equal(np.random.rand(3, 4), expected_rand)

        np.random.seed(7)
        expected_mask = [
            [False, True, False, True],
            [True, True, True, False],
            [False, False, True, True]
        ]
        mask = random_bool_mask(3, 4)
        np.testing.assert_array_equal(mask, expected_mask)

    def test_permute_systems(self):
        X = np.arange(1, 13).reshape(3, 4)
        Y = -np.arange(1, 13).reshape(3, 4)

        np.random.seed(7)
        expected_mask = [[False], [True], [False]]
        mask = random_bool_mask(3, 1)
        np.testing.assert_array_equal(mask, expected_mask)

        np.random.seed(7)
        expected_X = [
            [1, 2, 3, 4],
            [-5, -6, -7, -8],
            [9, 10, 11, 12]
        ]
        expected_Y = [
            [-1, -2, -3, -4],
            [5, 6, 7, 8],
            [-9, -10, -11, -12]
        ]
        X_p, Y_p = permute_systems(X, Y)
        np.testing.assert_array_equal(X_p, expected_X)
        np.testing.assert_array_equal(Y_p, expected_Y)
        np.testing.assert_array_equal(X, np.arange(1, 13).reshape(3, 4))
        np.testing.assert_array_equal(Y, -np.arange(1, 13).reshape(3, 4))

    def test_permute_inputs(self):
        X = np.arange(1, 13).reshape(3, 4)
        Y = -np.arange(1, 13).reshape(3, 4)

        np.random.seed(7)
        expected_mask = [[False, True, False, True]]
        mask = random_bool_mask(1, 4)
        np.testing.assert_array_equal(mask, expected_mask)

        np.random.seed(7)
        expected_X = [
            [1, -2, 3, -4],
            [5, -6, 7, -8],
            [9, -10, 11, -12]
        ]
        expected_Y = [
            [-1, 2, -3, 4],
            [-5, 6, -7, 8],
            [-9, 10, -11, 12]
        ]
        X_p, Y_p = permute_inputs(X, Y)
        np.testing.assert_array_equal(X_p, expected_X)
        np.testing.assert_array_equal(Y_p, expected_Y)
        np.testing.assert_array_equal(X, np.arange(1, 13).reshape(3, 4))
        np.testing.assert_array_equal(Y, -np.arange(1, 13).reshape(3, 4))

    def test_permute_both(self):
        X = np.arange(1, 13).reshape(3, 4)
        Y = -np.arange(1, 13).reshape(3, 4)

        np.random.seed(7)
        expected_mask = [
            [False, True, False, True],
            [True, True, True, False],
            [False, False, True, True]
        ]
        mask = random_bool_mask(3, 4)
        np.testing.assert_array_equal(mask, expected_mask)

        # The True values should swap and the original matrices should be unchanged
        np.random.seed(7)
        expected_X = [
            [1, -2, 3, -4],
            [-5, -6, -7, 8],
            [9, 10, -11, -12]
        ]
        expected_Y = [
            [-1, 2, -3, 4],
            [5, 6, 7, -8],
            [-9, -10, 11, 12]
        ]
        X_p, Y_p = permute_both(X, Y)
        np.testing.assert_array_equal(X_p, expected_X)
        np.testing.assert_array_equal(Y_p, expected_Y)
        np.testing.assert_array_equal(X, np.arange(1, 13).reshape(3, 4))
        np.testing.assert_array_equal(Y, -np.arange(1, 13).reshape(3, 4))

    def test_bootstrap_diff_test(self):
        # Regression test
        np.random.seed(12)
        X = np.random.random((9, 5))
        Y = np.random.random((9, 5))
        Z = np.random.random((9, 5))
        corr_func = functools.partial(global_corr, pearsonr)

        np.random.seed(2)
        assert bootstrap_diff_test(corr_func, X, Y, Z, bootstrap_system_sample, False) == 0.958
        np.random.seed(2)
        assert bootstrap_diff_test(corr_func, Y, X, Z, bootstrap_system_sample, False) == 0.042

    def test_permutation_diff_test(self):
        # Regression test
        np.random.seed(12)
        X = np.random.random((9, 5))
        Y = np.random.random((9, 5))
        Z = np.random.random((9, 5))
        corr_func = functools.partial(global_corr, pearsonr)

        np.random.seed(2)
        self.assertAlmostEqual(permutation_diff_test(corr_func, X, Y, Z, permute_both, False), 0.97002997002997, places=4)
        np.random.seed(2)
        self.assertAlmostEqual(permutation_diff_test(corr_func, Y, X, Z, permute_both, False), 0.030969030969030968, places=4)

    def test_williams_diff_test(self):
        # This test verifies that the output is the same as the psych package for
        # several different randomly generated inputs
        N, M = 9, 5
        corr_func = functools.partial(global_corr, pearsonr)

        np.random.seed(12)
        X = np.random.random((N, M))
        Y = np.random.random((N, M))
        Z = np.random.random((N, M))

        # These are used as input to r.test
        # effective_N = N * M
        # r12 = corr_func(X, Z)
        # r13 = corr_func(Y, Z)
        # r23 = corr_func(X, Y)

        # One tail
        expected_pvalue = 0.2716978
        actual_pvalue = williams_diff_test(corr_func, X, Y, Z, False)
        self.assertAlmostEqual(expected_pvalue, actual_pvalue, places=5)

        # The opposite order should produce 1-0.2716978. r.test does not do this and
        # will return 0.2716978 because it assumes that r12 > r13.
        actual_pvalue = williams_diff_test(corr_func, Y, X, Z, False)
        self.assertAlmostEqual(1.0 - expected_pvalue, actual_pvalue, places=5)

        # Two tails
        expected_pvalue = 0.5433956
        actual_pvalue = williams_diff_test(corr_func, X, Y, Z, True)
        self.assertAlmostEqual(expected_pvalue, actual_pvalue, places=5)

        # Should not matter the order for two tails
        actual_pvalue = williams_diff_test(corr_func, Y, X, Z, True)
        self.assertAlmostEqual(expected_pvalue, actual_pvalue, places=5)

        X = np.random.random((N, M))
        Y = np.random.random((N, M))
        Z = np.random.random((N, M))

        corr_func = functools.partial(system_level_corr, spearmanr)

        # These are used as input to r.test
        # effective_N = N
        # r12 = corr_func(X, Z)
        # r13 = corr_func(Y, Z)
        # r23 = corr_func(X, Y)

        # One tail
        # Since r12 < r13, r.test will only replicate this result with the reversed input order
        expected_pvalue = 0.4658712
        actual_pvalue = williams_diff_test(corr_func, Y, X, Z, False)
        self.assertAlmostEqual(expected_pvalue, actual_pvalue, places=5)

        # r.test would return the same result here, but we return 1.0 - expected
        actual_pvalue = williams_diff_test(corr_func, X, Y, Z, False)
        self.assertAlmostEqual(1.0 - expected_pvalue, actual_pvalue, places=5)

        # Two tails
        expected_pvalue = 0.9317423
        actual_pvalue = williams_diff_test(corr_func, X, Y, Z, True)
        self.assertAlmostEqual(expected_pvalue, actual_pvalue, places=5)

        # Order doesn't matter
        actual_pvalue = williams_diff_test(corr_func, Y, X, Z, True)
        self.assertAlmostEqual(expected_pvalue, actual_pvalue, places=5)

    def test_corr_diff_test(self):
        # Regression test
        np.random.seed(12)
        X = np.random.random((20, 10))
        Y = np.random.random((20, 10))
        Z = np.random.random((20, 10))
        corr_func = functools.partial(global_corr, pearsonr)

        # Ensure it's the same result going through bootstrap_diff_test and corr_diff_test
        np.random.seed(2)
        expected = bootstrap_diff_test(corr_func, X, Y, Z, bootstrap_system_sample, False)
        np.random.seed(2)
        assert corr_diff_test(corr_func, X, Y, Z, 'bootstrap-system', False) == expected

        np.random.seed(2)
        expected = bootstrap_diff_test(corr_func, X, Y, Z, bootstrap_input_sample, False)
        np.random.seed(2)
        assert corr_diff_test(corr_func, X, Y, Z, 'bootstrap-input', False) == expected

        np.random.seed(2)
        expected = bootstrap_diff_test(corr_func, X, Y, Z, bootstrap_both_sample, False)
        np.random.seed(2)
        assert corr_diff_test(corr_func, X, Y, Z, 'bootstrap-both', False) == expected

        # Ensure it's the same result going through permutation_diff_test and corr_diff_test
        np.random.seed(2)
        expected = permutation_diff_test(corr_func, X, Y, Z, permute_systems, False)
        np.random.seed(2)
        assert corr_diff_test(corr_func, X, Y, Z, 'permutation-system', False) == expected

        np.random.seed(2)
        expected = permutation_diff_test(corr_func, X, Y, Z, permute_inputs, False)
        np.random.seed(2)
        assert corr_diff_test(corr_func, X, Y, Z, 'permutation-input', False) == expected

        np.random.seed(2)
        expected = permutation_diff_test(corr_func, X, Y, Z, permute_both, False)
        np.random.seed(2)
        assert corr_diff_test(corr_func, X, Y, Z, 'permutation-both', False) == expected

        # None cases
        assert corr_diff_test(corr_func, X, Y, Z, 'none', False) is None
        assert corr_diff_test(corr_func, X, Y, Z, None, False) is None

        with self.assertRaises(Exception):
            corr_diff_test(corr_func, X, Y, Z, 'does-not-exist', False)

    def test_bonferroni_partial_conjunction_pvalue_test(self):
        # Tests against https://github.com/rtmdrr/replicability-analysis-NLP/blob/master/Replicability_Analysis.py
        pvalues = [0.168, 0.297, 0.357, 0.019, 0.218, 0.001]
        assert bonferroni_partial_conjunction_pvalue_test(pvalues, alpha=0.05) == (1, [5])
        assert bonferroni_partial_conjunction_pvalue_test(pvalues, alpha=0.10) == (2, [5, 3])
        assert bonferroni_partial_conjunction_pvalue_test(pvalues, alpha=0.70) == (6, [5, 3, 0, 4, 1, 2])


import os
import pytest
import unittest

from sacrerouge.commands.correlate import aggregate_metrics
from sacrerouge.data import Metrics
from sacrerouge.io import JsonlReader

_metrics_file_path = 'datasets/duc-tac/duc2006/v1.0/task1.metrics.jsonl'


class TestDUC2006SystemLevel(unittest.TestCase):
    @pytest.mark.skipif(not os.path.exists(_metrics_file_path), reason='DUC 2006 metrics file does not exist')
    def test_system_level(self):
        summary_level_metrics = JsonlReader(_metrics_file_path, Metrics).read()
        system_level_metrics = aggregate_metrics(summary_level_metrics)

        # Check a few metrics to make sure they are equal to what's in the NIST files
        # NISTeval2/ROUGE/rouge2.jk.m.avg
        assert system_level_metrics['C']['rouge-2_jk']['recall'] == pytest.approx(13.260, 1e-2)
        assert system_level_metrics['D']['rouge-2_jk']['recall'] == pytest.approx(12.380, 1e-2)
        assert system_level_metrics['B']['rouge-2_jk']['recall'] == pytest.approx(11.788, 1e-2)
        assert system_level_metrics['24']['rouge-2_jk']['recall'] == pytest.approx(9.558, 1e-2)
        assert system_level_metrics['15']['rouge-2_jk']['recall'] == pytest.approx(9.097, 1e-2)

        # NISTeval2/ROUGE/rougeSU4.jk.m.avg
        assert system_level_metrics['C']['rouge-su4_jk']['recall'] == pytest.approx(18.385, 1e-2)
        assert system_level_metrics['D']['rouge-su4_jk']['recall'] == pytest.approx(17.814, 1e-2)
        assert system_level_metrics['B']['rouge-su4_jk']['recall'] == pytest.approx(17.665, 1e-2)
        assert system_level_metrics['24']['rouge-su4_jk']['recall'] == pytest.approx(15.529, 1e-2)
        assert system_level_metrics['12']['rouge-su4_jk']['recall'] == pytest.approx(14.755, 1e-2)

        # NISTeval/responsiveness/avg_content
        assert system_level_metrics['D']['content_responsiveness'] == pytest.approx(4.9, 1e-2)
        assert system_level_metrics['C']['content_responsiveness'] == pytest.approx(4.9, 1e-2)
        assert system_level_metrics['B']['content_responsiveness'] == pytest.approx(4.9, 1e-2)
        assert system_level_metrics['27']['content_responsiveness'] == pytest.approx(3.08, 1e-2)
        assert system_level_metrics['23']['content_responsiveness'] == pytest.approx(3.0, 1e-2)

        # NISTeval/responsiveness/avg_overall
        assert system_level_metrics['E']['overall_responsiveness'] == pytest.approx(4.9, 1e-2)
        assert system_level_metrics['D']['overall_responsiveness'] == pytest.approx(4.9, 1e-2)
        assert system_level_metrics['I']['overall_responsiveness'] == pytest.approx(4.8, 1e-2)
        assert system_level_metrics['27']['overall_responsiveness'] == pytest.approx(2.84, 1e-2)
        assert system_level_metrics['23']['overall_responsiveness'] == pytest.approx(2.76, 1e-2)

        # NISTeval2/BE/simple.jk.m.hm.avg
        assert system_level_metrics['C']['rouge-be-hm_jk']['recall'] == pytest.approx(9.905, 1e-2)
        assert system_level_metrics['B']['rouge-be-hm_jk']['recall'] == pytest.approx(7.847, 1e-2)
        assert system_level_metrics['D']['rouge-be-hm_jk']['recall'] == pytest.approx(7.466, 1e-2)
        assert system_level_metrics['24']['rouge-be-hm_jk']['recall'] == pytest.approx(5.107, 1e-2)
        assert system_level_metrics['23']['rouge-be-hm_jk']['recall'] == pytest.approx(5.049, 1e-2)



import argparse
from overrides import overrides

from sacrerouge.datasets.multinews import setup
from sacrerouge.commands import DatasetSetupSubcommand


@DatasetSetupSubcommand.register('multinews')
class MultiNewsSubcommand(DatasetSetupSubcommand):
    @overrides
    def add_subparser(self, parser: argparse._SubParsersAction):
        description = 'Setup the Multi-News dataset'
        self.parser = parser.add_parser('multinews', description=description, help=description)
        self.parser.add_argument(
            'output_dir',
            type=str,
            help='The directory where the data should be written'
        )
        self.parser.add_argument(
            '--force',
            action='store_true',
            help='Force the raw data to be downloaded even if it already exists on file'
        )
        self.parser.set_defaults(subfunc=self.run)

    @overrides
    def run(self, args):
        setup.setup(args.output_dir, args.force)



import jsons

from sacrerouge.data.fields import Fields


class EvalInstance(object):
    def __init__(self,
                 instance_id: str,
                 summarizer_id: str,
                 summarizer_type: str,
                 fields: Fields) -> None:
        self.instance_id = instance_id
        self.summarizer_id = summarizer_id
        self.summarizer_type = summarizer_type
        self.fields = fields

    def __repr__(self) -> str:
        return jsons.dumps(self)



import os
import pytest

from sacrerouge.common.testing.metric_test_cases import ReferenceBasedMetricTestCase
from sacrerouge.common.testing.util import sacrerouge_command_exists
from sacrerouge.metrics import S3


@pytest.mark.skipif('S3_ENV' not in os.environ, reason='S3 python environment environment variable not set')
class TestS3(ReferenceBasedMetricTestCase):
    def test_s3(self):
        # This is a regression test, not necessarily a test for correctness
        metric = S3(environment_name=os.environ['S3_ENV'])
        expected_output = [
            {'s3': {'pyr': 0.32191771068281394, 'resp': 0.43112947213398156}},
            {'s3': {'pyr': 0.49038453876892796, 'resp': 0.5689795160671101}},
            {'s3': {'pyr': 0.47786455220076274, 'resp': 0.548172710896836}},
            {'s3': {'pyr': 0.3803879523604371, 'resp': 0.46652343758025827}},
            {'s3': {'pyr': 0.35703890425311635, 'resp': 0.43948176094166214}},
            {'s3': {'pyr': 0.39784433258660734, 'resp': 0.48528832257788146}},
            {'s3': {'pyr': 0.5483975017522531, 'resp': 0.6088455902978009}},
            {'s3': {'pyr': 0.5510649124151888, 'resp': 0.6052093249379357}},
            {'s3': {'pyr': 0.3891717207498385, 'resp': 0.48441896558584596}},
            {'s3': {'pyr': 0.5833618133817302, 'resp': 0.6298526245098648}},
            {'s3': {'pyr': 0.5346016112603917, 'resp': 0.593030815989284}},
            {'s3': {'pyr': 0.46100829186334746, 'resp': 0.5209557553762801}}
        ]
        super().assert_expected_output(metric, expected_output)

    def test_s3_order_invariant(self):
        metric = S3(environment_name=os.environ['S3_ENV'])
        self.assert_order_invariant(metric)

    def test_command_exists(self):
        assert sacrerouge_command_exists(['s3'])

    def test_setup_command_exists(self):
        assert sacrerouge_command_exists(['setup-metric', 's3'])


import lxml.html
import os
import tarfile
from collections import defaultdict
from typing import Dict, List

from sacrerouge.data import Metrics, MetricsDict
from sacrerouge.io import JsonlWriter


def load_summaries(summaries_tar: str, dirs: List[str]):
    summaries = defaultdict(dict)
    with tarfile.open(summaries_tar, 'r') as tar:
        for member in tar.getmembers():
            dirname = os.path.basename(os.path.dirname(member.name))
            if dirname in dirs and member.name.endswith('.html'):
                path = member.name.split('/')
                filename = path[-1]
                sections = filename.split('.')
                summarizer_id = sections[4]
                if summarizer_id.isalpha():
                    summarizer_type = 'reference'
                else:
                    summarizer_type = 'peer'

                # Some of these files appear to be html and others plain text
                text = tar.extractfile(member).read().decode(errors='replace')
                if text.startswith('<html>'):
                    tree = lxml.html.document_fromstring(text)
                    nodes = list(tree.xpath('//a[@id]'))
                    sentences = [node.text_content() for node in nodes]
                else:
                    sentences = text.splitlines()

                summary = {
                    'summarizer_id': summarizer_id,
                    'summarizer_type': summarizer_type,
                    'text': sentences
                }

                is_multi = sections[1] == 'M'
                if is_multi:
                    assert len(sections) == 6
                    cluster = sections[0].lower()
                    summaries[cluster][summarizer_id] = summary
                else:
                    assert len(sections) == 8
                    doc_id = '.'.join(sections[5:7])
                    summaries[doc_id][summarizer_id] = summary
    return summaries


def load_results_table(file_path: str, metrics):
    with open(file_path, 'r') as f:
        for line in f:
            columns = line.strip().split()
            assert len(columns) == 34, (len(columns), columns)
            if columns[1] == 'M':
                instance_id = columns[0].lower()
            else:
                instance_id = columns[2]

            summarizer_id = columns[8]

            summary_length = int(columns[3])
            # Length-10 summaries did not have quality questions
            if summary_length != 10:
                for i in range(12):
                    metrics[instance_id][summarizer_id]['peer_quality'][f'Q{i + 1}'] = int(columns[11 + i])

            metrics[instance_id][summarizer_id]['unmarked_units_related_to_subject'] = float(columns[23])
            metrics[instance_id][summarizer_id]['num_peer_units'] = int(columns[24])
            metrics[instance_id][summarizer_id]['num_marked_peer_units'] = int(columns[25])
            metrics[instance_id][summarizer_id]['num_unmarked_peer_units'] = int(columns[26])
            metrics[instance_id][summarizer_id]['num_model_units'] = int(columns[27])
            metrics[instance_id][summarizer_id]['coverage']['mean'] = float(columns[28])
            metrics[instance_id][summarizer_id]['coverage']['median'] = float(columns[29])
            metrics[instance_id][summarizer_id]['coverage']['std'] = float(columns[30])
            metrics[instance_id][summarizer_id]['length_adjusted_coverage']['mean'] = float(columns[31])
            metrics[instance_id][summarizer_id]['length_adjusted_coverage']['median'] = float(columns[32])
            metrics[instance_id][summarizer_id]['length_adjusted_coverage']['std'] = float(columns[33])


def load_responsiveness_table(file_path: str, metrics):
    with open(file_path, 'r') as f:
        for line in f:
            columns = line.strip().split()
            assert len(columns) == 7
            instance_id = columns[0].lower()
            summarizer_id = columns[2]
            score1 = int(columns[4])
            score2 = int(columns[6])
            metrics[instance_id][summarizer_id]['responsiveness'] = [score1, score2]


def load_usefulness_table(file_path: str, metrics):
    with open(file_path, 'r') as f:
        for line in f:
            columns = line.strip().split()
            assert len(columns) == 8
            instance_id = columns[1]
            # Annoying, the period was removed
            instance_id = instance_id[:-4] + '.' + instance_id[-4:]
            summarizer_id = columns[3]
            score1 = int(columns[5])
            score2 = int(columns[7])
            metrics[instance_id][summarizer_id]['usefulness'] = [score1, score2]


def get_references(summaries, instance_id, summarizer_id):
    summarizer_ids = list(summaries[instance_id].keys())
    reference_ids = list(filter(lambda sid: sid.isalpha(), summarizer_ids))

    references = []
    for reference_id in reference_ids:
        if summarizer_id == reference_id:
            continue
        references.append(summaries[instance_id][reference_id])
    return references


def save_metrics(summaries: Dict[str, Dict[str, List[str]]],
                 metrics: Dict[str, Dict[str, List[int]]],
                 task_name: str,
                 output_dir: str):
    with JsonlWriter(f'{output_dir}/{task_name}.summaries.jsonl') as out_summaries:
        with JsonlWriter(f'{output_dir}/{task_name}.metrics.jsonl') as out_metrics:
            for instance_id in sorted(summaries.keys()):
                for summarizer_id in summaries[instance_id].keys():
                    summary = summaries[instance_id][summarizer_id]
                    instance_metrics = metrics[instance_id][summarizer_id]
                    references = get_references(summaries, instance_id, summarizer_id)

                    out_summaries.write({
                        'instance_id': instance_id,
                        'summarizer_id': summarizer_id,
                        'summarizer_type': summary['summarizer_type'],
                        'summary': summary,
                        'references': references
                    })
                    out_metrics.write(Metrics(instance_id, summarizer_id, summary['summarizer_type'], instance_metrics))


def setup(data_root: str, output_dir: str) -> None:
    results_dir = f'{data_root}/scrapes/duc.nist.gov/past_duc/duc2003/results'
    main(results_dir, output_dir)


def main(results_dir, output_dir):
    summaries_tar = f'{results_dir}/detagged.duc2003.abstracts.tar.gz'
    task1_summaries = load_summaries(summaries_tar, ['peer1.1', 'peer6.1', 'peer7.1'])
    task2_summaries = load_summaries(summaries_tar, ['peer2.2', 'peer3.2', 'peer6.2', 'peer7.2'])
    task3_summaries = load_summaries(summaries_tar, ['peer2.3', 'peer3.3', 'peer6.3', 'peer7.3'])
    task4_summaries = load_summaries(summaries_tar, ['peer4.4', 'peer5.4', 'peer6.4', 'peer7.4'])

    metrics = defaultdict(lambda: defaultdict(MetricsDict))
    load_results_table(f'{results_dir}/original.results.corrected.improved.withPenaltyForExcess/short.results.table1',
                       metrics)

    task2_metrics = defaultdict(lambda: defaultdict(MetricsDict))
    load_results_table(f'{results_dir}/original.results.corrected.improved.withPenaltyForExcess/short.results.table2',
                       metrics)

    task3_metrics = defaultdict(lambda: defaultdict(MetricsDict))
    load_results_table(f'{results_dir}/original.results.corrected.improved.withPenaltyForExcess/short.results.table3',
                       metrics)

    task4_metrics = defaultdict(lambda: defaultdict(MetricsDict))
    load_results_table(f'{results_dir}/original.results.corrected.improved.withPenaltyForExcess/short.results.table4',
                       metrics)

    load_responsiveness_table(f'{results_dir}/results.table.R.paired.noheader', metrics)
    load_usefulness_table(f'{results_dir}/results.table.U.paired.noheader', metrics)

    save_metrics(task1_summaries, metrics, 'task1', output_dir)
    save_metrics(task2_summaries, metrics, 'task2', output_dir)
    save_metrics(task3_summaries, metrics, 'task3', output_dir)
    save_metrics(task4_summaries, metrics, 'task4', output_dir)



import argparse
import numpy as np
import os
from collections import defaultdict
from overrides import overrides
from subprocess import Popen
from typing import List

from sacrerouge.commands import MetricSetupSubcommand
from sacrerouge.common import DATA_ROOT
from sacrerouge.data import MetricsDict
from sacrerouge.data.types import ReferenceType, SummaryType
from sacrerouge.metrics import Metric, ReferenceBasedMetric

MOVERSCORE_INSTALLED = False

try:
    from moverscore_v2 import get_idf_dict, word_mover_score

    MOVERSCORE_INSTALLED = True

    @Metric.register('moverscore')
    class MoverScore(ReferenceBasedMetric):
        def __init__(self, moverscore_root: str = f'{DATA_ROOT}/metrics/MoverScore'):
            super().__init__()
            if not os.path.exists(moverscore_root):
                raise Exception(f'Path "{moverscore_root}" does not exist. Have you setup MoverScore?')
            self.stopwords = set(open(f'{moverscore_root}/stopwords.txt', 'r').read().strip().split())

        def _flatten_summary(self, summary: SummaryType) -> str:
            if isinstance(summary, list):
                return ' '.join(summary)
            return summary

        def _flatten_summaries(self, summaries_list: List[List[SummaryType]]) -> List[List[str]]:
            return [[self._flatten_summary(summary) for summary in summaries] for summaries in summaries_list]

        def _get_unique_summaries(self, summaries_list: List[List[str]]) -> List[str]:
            # Computing MoverScore requires calculating the IDF on the summary and
            # reference set. In sacreROUGE, the scoring functions could potentially
            # repeat references and summaries (e.g., if some of the scoring instances are
            # jackknifing examples), so we should not directly compute IDF on all the
            # summaries, but instead on the unique set of summaries
            unique_summaries = set([summary for summaries in summaries_list for summary in summaries])
            return list(unique_summaries)

        def _run(self,
                 summaries_list: List[List[SummaryType]],
                 references_list: List[List[SummaryType]]) -> List[List[MetricsDict]]:
            summaries_list = self._flatten_summaries(summaries_list)
            references_list = self._flatten_summaries(references_list)

            unique_summaries = self._get_unique_summaries(summaries_list)
            unique_references = self._get_unique_summaries(references_list)

            idf_dict_summaries = get_idf_dict(unique_summaries)
            idf_dict_references = get_idf_dict(unique_references)

            # Prepare the inputs into flat lists for faster processing. The
            # indices will keep track of which item the score belongs to
            indices = []
            input_summaries = []
            input_references = []
            for i, (summaries, references) in enumerate(zip(summaries_list, references_list)):
                for j, summary in enumerate(summaries):
                    for reference in references:
                        indices.append((i, j))
                        input_summaries.append(summary)
                        input_references.append(reference)

            # Score all of the data
            scores = word_mover_score(input_references, input_summaries,
                                      idf_dict_references, idf_dict_summaries,
                                      self.stopwords, n_gram=1, remove_subwords=True,
                                      batch_size=48)

            # Compute the mean over the references
            indices_to_scores = defaultdict(list)
            for pair, score in zip(indices, scores):
                indices_to_scores[pair].append(score)

            indices_to_score = {}
            for pair, scores in indices_to_scores.items():
                indices_to_score[pair] = np.mean(scores)

            # Put back into lists
            metrics_dict_lists = []
            for i in range(len(summaries_list)):
                metrics_dict_lists.append([])
                for j in range(len(summaries_list[i])):
                    metrics_dict_lists[-1].append(MetricsDict({
                        'MoverScore': indices_to_score[(i, j)]
                    }))
            return metrics_dict_lists

        @overrides
        def score_multi_all(self,
                            summaries_list: List[List[SummaryType]],
                            references_list: List[List[ReferenceType]]) -> List[List[MetricsDict]]:
            return self._run(summaries_list, references_list)

except ImportError:
    @Metric.register('moverscore')
    class MoverScore(Metric):
        @overrides
        def score_multi_all(self,
                            summaries_list: List[List[SummaryType]],
                            references_list: List[List[ReferenceType]]) -> List[List[MetricsDict]]:
            raise NotImplementedError('Error: "moverscore" python package is not installed')


@MetricSetupSubcommand.register('moverscore')
class MoverScoreSetupSubcommand(MetricSetupSubcommand):
    @overrides
    def add_subparser(self, parser: argparse._SubParsersAction):
        description = 'Setup the MoverScore metric'
        self.parser = parser.add_parser('moverscore', description=description, help=description)
        self.parser.set_defaults(subfunc=self.run)

    @overrides
    def run(self, args):
        commands = [
            f'mkdir -p {DATA_ROOT}/metrics/MoverScore',
            f'cd {DATA_ROOT}/metrics/MoverScore',
            f'wget https://raw.githubusercontent.com/AIPHES/emnlp19-moverscore/0c9380437706e59ba6510fb755d8836d633a9ea1/examples/stopwords.txt'
        ]
        command = ' && '.join(commands)

        process = Popen(command, shell=True)
        process.communicate()
        if process.returncode == 0:
            print('MoverScore data downloaded. Please install the "moverscore" pip package.')
        else:
            print('MoverScore setup failure')



import argparse
import json
import logging
import os
from collections import defaultdict
from overrides import overrides
from typing import Dict, List,  Union

from sacrerouge.commands import RootSubcommand
from sacrerouge.common.logging import prepare_global_logging
from sacrerouge.stats import partial_conjunction_pvalue_test

logger = logging.getLogger(__name__)


def load_pvalue_dicts(pvalue_json_files_or_dicts: Union[List[str], List[Dict]],
                      names: List[str]) -> Dict[str, Dict[str, Dict[str, float]]]:
    assert len(pvalue_json_files_or_dicts) == len(names)
    pvalues_list = []
    for json_file_or_dict in pvalue_json_files_or_dicts:
        if isinstance(json_file_or_dict, str):
            pvalues = json.load(open(json_file_or_dict, 'r'))
        else:
            pvalues = json_file_or_dict
        pvalues_list.append(pvalues)

    # Ensure they all have the same levels
    levels = set()
    for level in ['summary_level', 'system_level', 'global']:
        if level in pvalues_list[0]:
            levels.add(level)
    for pvalues in pvalues_list:
        for level in levels:
            assert level in pvalues

    pvalue_dict = defaultdict(lambda: defaultdict(dict))
    for name, pvalues in zip(names, pvalues_list):
        for level in levels:
            for coef in ['pearson', 'spearman', 'kendall']:
                pvalue_dict[level][name][coef] = pvalues[level][coef]['pvalue']
    return pvalue_dict


def run_partial_conjunction_pvalue_test(method: str,
                                        pvalues_dict: Dict[str, Dict[str, float]],
                                        alpha: float = 0.05) -> float:
    results = {}
    for coef in ['pearson', 'spearman', 'kendall']:
        metrics = []
        pvalues = []
        for metric, ps in pvalues_dict.items():
            metrics.append(metric)
            pvalues.append(ps[coef])

        k, significant_indices = partial_conjunction_pvalue_test(method, pvalues, alpha=alpha)
        significant_metrics = [metrics[i] for i in significant_indices]
        results[coef] = {
            'k': k,
            'significant': significant_metrics
        }
    return results


def run_all_partial_conjunction_pvalue_test(method: str,
                                            pvalue_json_files_or_dicts: Union[List[str], List[Dict]],
                                            names: List[str],
                                            alpha: float = 0.05) -> Dict:
    pvalue_dict = load_pvalue_dicts(pvalue_json_files_or_dicts, names)
    results = {
        'method': method,
        'names': names,
        'alpha': alpha
    }
    for level in pvalue_dict.keys():
        results[level] = run_partial_conjunction_pvalue_test(method, pvalue_dict[level], alpha=alpha)
    return results


@RootSubcommand.register('partial-conjunction-test')
class PartialConjunctionTestSubcommand(RootSubcommand):
    @overrides
    def add_subparser(self, parser: argparse._SubParsersAction):
        description = 'Run a partial conjunction test on a set of p-values'
        self.parser = parser.add_parser('partial-conjunction-test', description=description, help=description)
        self.parser.add_argument(
            '--pvalue-json-files',
            nargs='+',
            help='The json files with the p-values to run the partial conjunction p-value test on',
            required=True
        )
        self.parser.add_argument(
            '--names',
            nargs='+',
            type=str,
            help='The names for the correspond pvalue json files',
            required=True
        )
        self.parser.add_argument(
            '--method',
            choices=['bonferroni'],
            default='bonferroni',
            help='The partial conjunction p-value method to use'
        )
        self.parser.add_argument(
            '--alpha',
            type=float,
            default=0.05,
            help='The signifiance level of the test'
        )
        self.parser.add_argument(
            '--output-file',
            type=str,
            help='The json output file which will contain the final correlations'
        )
        self.parser.add_argument(
            '--log-file',
            type=str,
            help='The file where the log should be written'
        )
        self.parser.add_argument(
            '--silent',
            action='store_true',
            help='Controls whether the log should be written to stdout'
        )
        self.parser.set_defaults(func=self.run)

    def run(self, args):
        prepare_global_logging(file_path=args.log_file, silent=args.silent)

        results = run_all_partial_conjunction_pvalue_test(args.method, args.pvalue_json_files, args.names,
                                                          alpha=args.alpha)

        if args.output_file:
            dirname = os.path.dirname(args.output_file)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
            with open(args.output_file, 'w') as out:
                out.write(json.dumps(results, indent=2))

        if not args.silent:
            logger.info(json.dumps(results, indent=2))


from sacrerouge.metrics.metric import DocumentBasedMetric, Metric, ReferenceBasedMetric, ReferenceFreeMetric, SummaryBasedMetric
from sacrerouge.metrics.apes import APES
from sacrerouge.metrics.autosummeng import AutoSummENG
from sacrerouge.metrics.bertscore import BertScore
from sacrerouge.metrics.bewte import BEwTE
from sacrerouge.metrics.blanc import Blanc
from sacrerouge.metrics.bleu import SentBleu
from sacrerouge.metrics.bleurt import Bleurt
from sacrerouge.metrics.decomposed_rouge import DecomposedRouge
from sacrerouge.metrics.meteor import Meteor
from sacrerouge.metrics.moverscore import MoverScore
from sacrerouge.metrics.pyramid_score import PyramidScore
from sacrerouge.metrics.pyreval import PyrEval
from sacrerouge.metrics.python_rouge import PythonRouge
from sacrerouge.metrics.qaeval import QAEval
from sacrerouge.metrics.rouge import Rouge
from sacrerouge.metrics.s3 import S3
from sacrerouge.metrics.simetrix import SIMetrix
from sacrerouge.metrics.sumqe import SumQE
from sacrerouge.metrics.supert import SUPERT



import logging
from sacrebleu import sentence_bleu
from typing import List

from sacrerouge.common.util import flatten
from sacrerouge.data import MetricsDict
from sacrerouge.data.types import ReferenceType
from sacrerouge.data.types import SummaryType
from sacrerouge.metrics import Metric, ReferenceBasedMetric

logger = logging.getLogger(__name__)


@Metric.register('sent-bleu')
class SentBleu(ReferenceBasedMetric):
    def __init__(self, **kwargs) -> None:
        """
        Args:
            **kwargs: The kwargs that will be passed to `sacrebleu.sentence_bleu`. See
                the `sacrebleu` documentation for details.
        """
        super().__init__()
        self.kwargs = kwargs

    def score_multi_all(
        self,
        summaries_list: List[List[SummaryType]],
        references_list: List[List[ReferenceType]],
        **kwargs,
    ) -> List[List[MetricsDict]]:
        scores_list = []
        for summaries, references in zip(summaries_list, references_list):
            references = [flatten(reference) for reference in references]
            scores_list.append([])
            for summary in summaries:
                summary = flatten(summary)
                score = sentence_bleu(summary, references, **self.kwargs)
                scores_list[-1].append(MetricsDict({'sent-bleu': score.score}))
        return scores_list


from spacy.tokens import Token
from typing import Dict, List, Tuple

from sacrerouge.data import MetricsDict
from sacrerouge.metrics.decomposed_rouge.categorizers.util import calculate_maximum_matching


class Categorizer(object):
    def __init__(self, name: str) -> None:
        self.name = name

    def is_match(self, token1: Token, token2: Token) -> bool:
        raise NotImplementedError

    def is_candidate(self, token: Token) -> bool:
        raise NotImplementedError

    def select_matches(self,
                       summary_tokens: List[Token],
                       reference_tokens: List[Token],
                       matches: List[Tuple[int, int]],
                       intersection: int):
        common_matches = []
        for i, j in matches:
            summary_token = summary_tokens[i]
            reference_token = reference_tokens[j]
            if self.is_match(summary_token, reference_token):
                common_matches.append((i, j, 1.0))

        num_matches = calculate_maximum_matching(common_matches)
        num_summary_tokens = sum(1 for token in summary_tokens if self.is_candidate(token))
        num_reference_tokens = sum(1 for token in reference_tokens if self.is_candidate(token))

        precision = num_matches / num_summary_tokens * 100 if num_summary_tokens > 0 else 0
        recall = num_matches / num_reference_tokens * 100 if num_reference_tokens > 0 else 0
        f1 = (2 * precision * recall) / (precision + recall) if (precision + recall) > 0 else 0

        contribution = num_matches / intersection * 100 if len(matches) > 0 else 0

        return MetricsDict({
            self.name: {
                'precision': precision,
                'recall': recall,
                'f1': f1,
                'contribution': contribution
            }
        })


class TupleCategorizer(Categorizer):
    def get_tuples(self, tokens: List[Token]) -> List[Dict[str, int]]:
        raise NotImplementedError

    def select_matches(self,
                       summary_tokens: List[Token],
                       reference_tokens: List[Token],
                       matches: List[Tuple[int, int]],
                       intersection: int):
        summary_tuples = self.get_tuples(summary_tokens)
        reference_tuples = self.get_tuples(reference_tokens)
        matches = set(matches)

        # Figure out the list of tuple-level matches based on whether the tuples
        # match each other completely
        tuple_matches = []
        for s_i, summary_tuple in enumerate(summary_tuples):
            for r_j, reference_tuple in enumerate(reference_tuples):
                assert len(summary_tuple) == len(reference_tuple)
                # See if each component of these two tuples can be aligned
                matched = True
                for key, i in summary_tuple.items():
                    if key not in reference_tuple:
                        matched = False
                        break
                    j = reference_tuple[key]
                    if (i, j) not in matches:
                        matched = False
                        break

                if matched:
                    tuple_matches.append((s_i, r_j, len(summary_tuple)))

        # Calculate the weight of the matched tuples, only allowing each tuple to be matched once.
        # The tuples form an equivalence class, so it doesn't matter exactly what match we use. This is
        # equivalent to calculating the size of the maximum matching in a bipartite graph where the
        # two disjoint sets of vertices are the summary and reference tuples, and an edge exists between
        # them if they were matched
        total_weight = calculate_maximum_matching(tuple_matches)

        summary_tuples_weight = sum(len(tup) for tup in summary_tuples)
        reference_tuples_weight = sum(len(tup) for tup in reference_tuples)

        precision = total_weight / summary_tuples_weight * 100 if summary_tuples_weight > 0 else 0
        recall = total_weight / reference_tuples_weight * 100 if reference_tuples_weight > 0 else 0
        f1 = (2 * precision * recall) / (precision + recall) if (precision + recall) > 0 else 0

        contribution = total_weight / intersection * 100 if len(matches) > 0 else 0

        return MetricsDict({
            self.name: {
                'precision': precision,
                'recall': recall,
                'f1': f1,
                'contribution': contribution
            }
        })


import os
import pytest
import unittest

from sacrerouge.commands.correlate import aggregate_metrics
from sacrerouge.data import Metrics
from sacrerouge.io import JsonlReader

_task1_metrics_file_path = 'datasets/duc-tac/duc2007/v1.0/task1.metrics.jsonl'
_task2_metrics_file_path = 'datasets/duc-tac/duc2007/v1.0/task2.A-B-C.metrics.jsonl'


class TestDUC2007SystemLevel(unittest.TestCase):
    @pytest.mark.skipif(not os.path.exists(_task1_metrics_file_path), reason='DUC 2007 task 1 metrics file does not exist')
    def test_task1_system_level(self):
        summary_level_metrics = JsonlReader(_task1_metrics_file_path, Metrics).read()
        system_level_metrics = aggregate_metrics(summary_level_metrics)

        # Check a few metrics to make sure they are equal to what's in the NIST files
        # mainEval/ROUGE/rouge2.jk.m.avg
        assert system_level_metrics['D']['rouge-2_jk']['recall'] == pytest.approx(17.528, 1e-2)
        assert system_level_metrics['C']['rouge-2_jk']['recall'] == pytest.approx(15.055, 1e-2)
        assert system_level_metrics['B']['rouge-2_jk']['recall'] == pytest.approx(13.992, 1e-2)
        assert system_level_metrics['15']['rouge-2_jk']['recall'] == pytest.approx(12.448, 1e-2)
        assert system_level_metrics['29']['rouge-2_jk']['recall'] == pytest.approx(12.028, 1e-2)

        # mainEval/ROUGE/rougeSU4.jk.m.avg
        assert system_level_metrics['D']['rouge-su4_jk']['recall'] == pytest.approx(21.892, 1e-2)
        assert system_level_metrics['C']['rouge-su4_jk']['recall'] == pytest.approx(19.921, 1e-2)
        assert system_level_metrics['E']['rouge-su4_jk']['recall'] == pytest.approx(19.396, 1e-2)
        assert system_level_metrics['15']['rouge-su4_jk']['recall'] == pytest.approx(17.711, 1e-2)
        assert system_level_metrics['24']['rouge-su4_jk']['recall'] == pytest.approx(17.593, 1e-2)

        # mainEval/manual/avg_content
        assert system_level_metrics['D']['content_responsiveness'] == pytest.approx(4.944, 1e-2)
        assert system_level_metrics['I']['content_responsiveness'] == pytest.approx(4.889, 1e-2)
        assert system_level_metrics['G']['content_responsiveness'] == pytest.approx(4.889, 1e-2)
        assert system_level_metrics['4']['content_responsiveness'] == pytest.approx(3.400, 1e-2)
        assert system_level_metrics['23']['content_responsiveness'] == pytest.approx(3.311, 1e-2)

        # mainEval/BE/simple.jk.m.hm.avg
        assert system_level_metrics['D']['rouge-be-hm_jk']['recall'] == pytest.approx(12.284, 1e-2)
        assert system_level_metrics['C']['rouge-be-hm_jk']['recall'] == pytest.approx(9.593, 1e-2)
        assert system_level_metrics['B']['rouge-be-hm_jk']['recall'] == pytest.approx(9.146, 1e-2)
        assert system_level_metrics['15']['rouge-be-hm_jk']['recall'] == pytest.approx(6.632, 1e-2)
        assert system_level_metrics['24']['rouge-be-hm_jk']['recall'] == pytest.approx(6.577, 1e-2)

    @pytest.mark.skipif(not os.path.exists(_task2_metrics_file_path), reason='DUC 2007 task 2 metrics file does not exist')
    def test_task2_system_level(self):
        summary_level_metrics = JsonlReader(_task2_metrics_file_path, Metrics).read()
        system_level_metrics = aggregate_metrics(summary_level_metrics)

        # Check a few metrics to make sure they are equal to what's in the NIST files
        # updateEval/ROUGE/rouge2.jk.m.avg
        assert system_level_metrics['D']['rouge-2_jk']['recall'] == pytest.approx(14.499, 1e-2)
        assert system_level_metrics['C']['rouge-2_jk']['recall'] == pytest.approx(14.330, 1e-2)
        assert system_level_metrics['G']['rouge-2_jk']['recall'] == pytest.approx(13.942, 1e-2)
        assert system_level_metrics['40']['rouge-2_jk']['recall'] == pytest.approx(11.189, 1e-2)
        assert system_level_metrics['55']['rouge-2_jk']['recall'] == pytest.approx(9.851, 1e-2)

        # updateEval/ROUGE/rougeSU4.jk.m.avg
        assert system_level_metrics['D']['rouge-su4_jk']['recall'] == pytest.approx(17.998, 1e-2)
        assert system_level_metrics['C']['rouge-su4_jk']['recall'] == pytest.approx(17.923, 1e-2)
        assert system_level_metrics['E']['rouge-su4_jk']['recall'] == pytest.approx(17.689, 1e-2)
        assert system_level_metrics['40']['rouge-su4_jk']['recall'] == pytest.approx(14.306, 1e-2)
        assert system_level_metrics['44']['rouge-su4_jk']['recall'] == pytest.approx(13.607, 1e-2)

        # updateEval/manual/Responsiveness/avg_content.all
        assert system_level_metrics['D']['content_responsiveness'] == pytest.approx(4.833, 1e-2)
        assert system_level_metrics['C']['content_responsiveness'] == pytest.approx(4.833, 1e-2)
        assert system_level_metrics['E']['content_responsiveness'] == pytest.approx(4.750, 1e-2)
        assert system_level_metrics['40']['content_responsiveness'] == pytest.approx(2.967, 1e-2)
        assert system_level_metrics['36']['content_responsiveness'] == pytest.approx(2.800, 1e-2)

        # updateEval/BE/simple.jk.m.hm.avg
        assert system_level_metrics['D']['rouge-be-hm_jk']['recall'] == pytest.approx(10.687, 1e-2)
        assert system_level_metrics['C']['rouge-be-hm_jk']['recall'] == pytest.approx(10.214, 1e-2)
        assert system_level_metrics['E']['rouge-be-hm_jk']['recall'] == pytest.approx(10.177, 1e-2)
        assert system_level_metrics['40']['rouge-be-hm_jk']['recall'] == pytest.approx(7.219, 1e-2)
        assert system_level_metrics['44']['rouge-be-hm_jk']['recall'] == pytest.approx(5.544, 1e-2)



import json
import unittest
from collections import defaultdict
from subprocess import PIPE, Popen
from typing import Dict, List

from sacrerouge.common import TemporaryDirectory
from sacrerouge.common.testing import FIXTURES_ROOT, MULTILING_SUMMARIES
from sacrerouge.data import Metrics
from sacrerouge.io import JsonlReader

_config_file_path = f'{FIXTURES_ROOT}/configs/evaluate.json'
_numeric_config_file_path = f'{FIXTURES_ROOT}/configs/evaluate-numeric.json'


class TestMetricCommand(unittest.TestCase):
    def _check_macro_default(self, macro: Dict) -> None:
        assert macro['metrics'] == {
            'python-rouge-1': {
                'precision': 47.288299019868255,
                'recall': 46.36412967789611,
                'f1': 46.734012277932806
            },
            'python-rouge-2': {
                'precision': 16.625951207280107,
                'recall': 16.30045341659763,
                'f1': 16.430312150593775
            }
        }

    def _check_micro_list_default(self, micro_list: List[Metrics]) -> None:
        instances = JsonlReader(MULTILING_SUMMARIES).read()

        assert len(micro_list) == len(instances)
        for micro, instance in zip(micro_list, instances):
            assert micro.instance_id == instance['instance_id']
            assert micro.summarizer_id == instance['summarizer_id']
            assert micro.summarizer_type == instance['summarizer_type']

        assert micro_list[0].metrics == {
            'python-rouge-1': {
                'precision': 41.699867197875164,
                'recall': 40.516129032258064,
                'f1': 41.09947643979057
            },
            'python-rouge-2': {
                'precision': 10.533333333333333,
                'recall': 10.233160621761659,
                'f1': 10.38107752956636
            }
        }
        assert micro_list[11].metrics == {
            'python-rouge-1': {
                'precision': 45.34412955465587,
                'recall': 44.71057884231537,
                'f1': 45.0251256281407
            },
            'python-rouge-2': {
                'precision': 15.24390243902439,
                'recall': 15.030060120240481,
                'f1': 15.136226034308779
            }
        }

    def _check_macro_arguments(self, macro: Dict) -> None:
        assert macro['metrics'] == {
            'python-rouge-3': {
              'precision': 3.8408974398855293,
              'recall': 3.879200450421004,
              'f1': 3.8496095259110024
            },
            'python-rouge-l': {
              'precision': 27.622060602200875,
              'recall': 27.547088418712146,
              'f1': 27.50970881170657
            }
        }

    def _check_micro_list_arguments(self, micro_list: List[Metrics]) -> None:
        instances = JsonlReader(MULTILING_SUMMARIES).read()

        assert len(micro_list) == len(instances)
        for micro, instance in zip(micro_list, instances):
            assert micro.instance_id == instance['instance_id']
            assert micro.summarizer_id == instance['summarizer_id']
            assert micro.summarizer_type == instance['summarizer_type']

        assert micro_list[0].metrics == {
            'python-rouge-3': {
                'precision': 3.0952380952380953,
                'recall': 3.110047846889952,
                'f1': 3.1026252983293556
            },
            'python-rouge-l': {
                'precision': 20.657276995305164,
                'recall': 20.754716981132077,
                'f1': 20.705882352941174
            }
        }
        assert micro_list[11].metrics == {
            'python-rouge-3': {
                'precision': 4.273504273504273,
                'recall': 3.816793893129771,
                'f1': 4.03225806451613
            },
                'python-rouge-l': {
                'precision': 28.15126050420168,
                'recall': 25.18796992481203,
                'f1': 26.58730158730159
            }
        }

    def test_evaluate_default(self):
        # I manually ran evaluate with these parameters and this method checks to make sure
        # those values are equal to the output here
        with TemporaryDirectory() as temp_dir:
            macro_file = f'{temp_dir}/macro.json'
            micro_file = f'{temp_dir}/micro.jsonl'
            command = [
                'python', '-m', 'sacrerouge', 'python-rouge', 'evaluate',
                '--macro-output-json', macro_file,
                '--micro-output-jsonl', micro_file,
                '--dataset-reader', 'reference-based',
                '--input-files', MULTILING_SUMMARIES,
                '--silent'
            ]

            process = Popen(command, stdout=PIPE, stderr=PIPE)
            process.communicate()

            macro_metrics = json.load(open(macro_file, 'r'))
            micro_metrics_list = JsonlReader(micro_file, Metrics).read()

            self._check_macro_default(macro_metrics)
            self._check_micro_list_default(micro_metrics_list)

    def test_evaluate_arguments(self):
        # I manually ran evaluate with these parameters and this method checks to make sure
        # those values are equal to the output here
        with TemporaryDirectory() as temp_dir:
            macro_file = f'{temp_dir}/macro.json'
            micro_file = f'{temp_dir}/micro.jsonl'
            command = [
                'python', '-m', 'sacrerouge', 'python-rouge', 'evaluate',
                '--macro-output-json', macro_file,
                '--micro-output-jsonl', micro_file,
                '--ngram_orders', '[3]',
                '--use_porter_stemmer', 'false',
                '--remove_stopwords', 'true',
                '--compute_rouge_l', 'true',
                '--dataset-reader', 'reference-based',
                '--input-files', MULTILING_SUMMARIES,
                '--silent'
            ]

            process = Popen(command, stdout=PIPE, stderr=PIPE)
            process.communicate()

            macro_metrics = json.load(open(macro_file, 'r'))
            micro_metrics_list = JsonlReader(micro_file, Metrics).read()

            self._check_macro_arguments(macro_metrics)
            self._check_micro_list_arguments(micro_metrics_list)

    def test_score_default(self):
        with TemporaryDirectory() as temp_dir:
            # I manually ran the scoring function with these parameters, and this test makes sure
            # those are equal to the output here
            output_file = f'{temp_dir}/metrics.jsonl'
            command = [
                'python', '-m', 'sacrerouge', 'python-rouge', 'score',
                '--output-jsonl', output_file,
                '--dataset-reader', 'reference-based',
                '--input-files', MULTILING_SUMMARIES,
                '--silent'
            ]

            process = Popen(command, stdout=PIPE, stderr=PIPE)
            process.communicate()

            instances = JsonlReader(MULTILING_SUMMARIES).read()
            metrics_list = JsonlReader(output_file, Metrics).read()
            metrics_dicts = defaultdict(dict)

            assert len(instances) == len(metrics_list)
            for instance, metrics in zip(instances, metrics_list):
                assert metrics.instance_id == instance['instance_id']
                assert metrics.summarizer_id == instance['summarizer_id']
                assert metrics.summarizer_type == instance['summarizer_type']
                metrics_dicts[metrics.instance_id][metrics.summarizer_id] = metrics

                if metrics.summarizer_type == 'reference':
                    assert 'python-rouge-1_jk' in metrics.metrics
                    assert 'python-rouge-2_jk' in metrics.metrics
                else:
                    assert 'python-rouge-1' in metrics.metrics
                    assert 'python-rouge-2' in metrics.metrics
                    assert 'python-rouge-1_jk' in metrics.metrics
                    assert 'python-rouge-2_jk' in metrics.metrics

            assert metrics_dicts['M000']['1'].metrics == {
                'python-rouge-1': {
                  'precision': 41.699867197875164,
                  'recall': 40.516129032258064,
                  'f1': 41.09947643979057
                },
                'python-rouge-2': {
                  'precision': 10.533333333333333,
                  'recall': 10.233160621761659,
                  'f1': 10.38107752956636
                },
                'python-rouge-1_jk': {
                  'precision': 41.699867197875164,
                  'recall': 40.514662613316766,
                  'f1': 41.098355761265616
                },
                'python-rouge-2_jk': {
                  'precision': 10.533333333333333,
                  'recall': 10.226158358122346,
                  'f1': 10.3773782079838
                }
            }

            assert metrics_dicts['M001']['B'].metrics == {
                'python-rouge-1_jk': {
                  'precision': 51.59362549800797,
                  'recall': 51.18577075098815,
                  'f1': 51.3888888888889
                },
                'python-rouge-2_jk': {
                  'precision': 20.4,
                  'recall': 20.238095238095237,
                  'f1': 20.318725099601597
                }
            }

    def test_score_arguments(self):
        with TemporaryDirectory() as temp_dir:
            # I manually ran the scoring function with these parameters, and this test makes sure
            # those are equal to the output here
            output_file = f'{temp_dir}/metrics.jsonl'
            command = [
                'python', '-m', 'sacrerouge', 'python-rouge', 'score',
                '--output-jsonl', output_file,
                '--dataset-reader', 'reference-based',
                '--input-files', MULTILING_SUMMARIES,
                '--ngram_orders', '[3]',
                '--use_porter_stemmer', 'false',
                '--remove_stopwords', 'true',
                '--compute_rouge_l', 'true',
                '--dataset-reader', 'reference-based',
                '--silent'
            ]

            process = Popen(command, stdout=PIPE, stderr=PIPE)
            process.communicate()

            instances = JsonlReader(MULTILING_SUMMARIES).read()
            metrics_list = JsonlReader(output_file, Metrics).read()
            metrics_dicts = defaultdict(dict)

            assert len(instances) == len(metrics_list)
            for instance, metrics in zip(instances, metrics_list):
                assert metrics.instance_id == instance['instance_id']
                assert metrics.summarizer_id == instance['summarizer_id']
                assert metrics.summarizer_type == instance['summarizer_type']
                metrics_dicts[metrics.instance_id][metrics.summarizer_id] = metrics

                if metrics.summarizer_type == 'reference':
                    assert 'python-rouge-3_jk' in metrics.metrics
                    assert 'python-rouge-l_jk' in metrics.metrics
                else:
                    assert 'python-rouge-3' in metrics.metrics
                    assert 'python-rouge-l' in metrics.metrics
                    assert 'python-rouge-3_jk' in metrics.metrics
                    assert 'python-rouge-l_jk' in metrics.metrics

            assert metrics_dicts['M000']['1'].metrics == {
                'python-rouge-3': {
                    'precision': 3.0952380952380953,
                    'recall': 3.110047846889952,
                    'f1': 3.1026252983293556
                },
                'python-rouge-l': {
                    'precision': 20.657276995305164,
                    'recall': 20.754716981132077,
                    'f1': 20.705882352941174
                },
                'python-rouge-3_jk': {
                    'precision': 3.095238095238095,
                    'recall': 3.073768703921825,
                    'f1': 3.0843425372732653
                },
                'python-rouge-l_jk': {
                    'precision': 20.657276995305164,
                    'recall': 20.72789236755767,
                    'f1': 20.69018908745478
                }
            }

            assert metrics_dicts['M001']['B'].metrics == {
                'python-rouge-3_jk': {
                    'precision': 3.75,
                    'recall': 3.4615384615384617,
                    'f1': 3.6
                },
                'python-rouge-l_jk': {
                    'precision': 33.60655737704918,
                    'recall': 31.060606060606062,
                    'f1': 32.28346456692913
                }
            }


import logging
import os
import sys


def prepare_global_logging(file_path: str = None, silent: bool = False) -> None:
    """
    Sets up the logger. If `file_path` is not `None`, the log is written to that file. If
    `silent` is not `True`, the log will also be written to stdout.
    """
    handlers = []
    if file_path is not None:
        dirname = os.path.dirname(file_path)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        handlers.append(logging.FileHandler(file_path))
    if not silent:
        handlers.append(logging.StreamHandler(sys.stdout))

    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
        handlers=handlers
    )
